# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
triavg.py

Contains a class to define the the POWDER averaging algorithm for NMR spectra,
as described in:

D. W. Alderman, Mark S. Solum, and David M. Grant
Methods for analyzing spectroscopic line shapes. NMR solid powder patterns
[J. Chern. Phys. 84 (7), 1 April 1986]
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from soprano.calculate.powder.powder import PowderScheme


class TriAvg(PowderScheme):

    def get_orient_points(self, N):
        """
        Generate and return the TriAvg angles (in the form of direction cosines),
        weights, and triangles.

        | Args:
        |   N (int): number of divisions on each side of the octahedron used to
        |            generate angles. Higher numbers lead to greater precision.

        | Returns:
        |   points, weights, tris (np.ndarray): arrays containing respectively the
        |                                       direction cosines for each
        |                                       orientation, the weights, and the
        |                                       triangles (in form of triplets of
        |                                       indices of the first array)

        """

        # Use the POWDER algorithm
        # First, generate the positive octant, by row
        points = [np.arange(N-z+1) for z in range(N+1)]
        points = np.concatenate([list(zip(p, N-z-p, [z]*(N-z+1)))
                                 for z, p in enumerate(points)])*1.0/N

        def z_i(z):
            return int(z*N+1.5*z-z**2/2.0)

        tris = np.array([[x, x+1, x+(N-z+1)]
                         for z in range(N) for x in range(z_i(z), z_i(z+1)-1)] +
                        [[x, x+(N-z), x+(N-z+1)]
                         for z in range(N) for x in range(z_i(z)+1, z_i(z+1)-1)])

        # Point weights
        dists = np.linalg.norm(points, axis=1)
        points /= dists[:, None]
        weights = dists**-3.0

        # Repeat on as many octants as needed
        if self.mode == 'octant':
            ranges = [[1], [1], [1]]
        elif self.mode == 'hemisphere':
            ranges = [[1, -1], [1, -1], [1]]
        elif self.mode == 'sphere':
            ranges = [[1, -1], [1, -1], [1, -1]]
        else:
            raise ValueError("Invalid mode passed to powder_alg")

        octants = np.array(np.meshgrid(*ranges)).T.reshape((-1, 3))

        points = (points[None, :, :]*octants[:, None, :]).reshape((-1, 3))
        weights = np.tile(weights, len(octants))
        tris = np.concatenate([tris]*len(octants)) + \
            np.repeat(np.arange(len(octants))*((N+2)*(N+1))/2,
                      len(tris))[:, None]

        # Some points are duplicated though. Remove them.
        ncols = points.shape[1]
        dtype = points.dtype.descr * ncols
        struct = points.view(dtype)
        uniq, uniq_i, uniq_inv = np.unique(struct, return_index=True,
                                           return_inverse=True)
        points = uniq.view(points.dtype).reshape(-1, ncols)

        # Remap triangles
        weights = weights[uniq_i]
        tris = uniq_inv[tris.astype(int)]

        return points, weights, tris

    def get_orient_angles(self, N):
        """
        Generate and return the TriAvg angles (in the form of angles in radians),
        weights, and triangles.

        | Args:
        |   N (int): number of divisions on each side of the octahedron used to
        |            generate angles. Higher numbers lead to greater precision.

        | Returns:
        |   points, weights, tris (np.ndarray): arrays containing respectively the
        |                                       direction cosines for each
        |                                       orientation, the weights, and the
        |                                       triangles (in form of triplets of
        |                                       indices of the first array)

        """

        points, weights, tris = self.get_orient_points()

        theta = np.arccos(points[:, 2])
        phi = np.arctan2(points[:, 1], points[:, 0])

        return np.array([theta, phi]).T, weights, tris

    def get_orient_trig(self, N):
        """
        Generate and return the TriAvg angles (in the form of trigonometric fuctions),
        weights, and triangles.

        | Args:
        |   N (int): number of divisions on each side of the octahedron used to
        |            generate angles. Higher numbers lead to greater precision.

        | Returns:
        |   points, weights, tris (np.ndarray): arrays containing respectively the
        |                                       direction cosines for each
        |                                       orientation, the weights, and the
        |                                       triangles (in form of triplets of
        |                                       indices of the first array)

        """

        points, weights, tris = self.get_orient_points()

        ct = points[:, 2]
        st = (1-ct**2)**0.5
        cp = points[:, 0]/st
        sp = points[:, 1]/st

        return np.array([ct, st, cp, sp]).T, weights, tris

    def average(self, x, y, weights, tris):

        triy = np.sort(y[tris], axis=1)
        triweights = np.average(weights[tris], axis=1)
        y = np.zeros(x.shape)

        dx = np.array([-x[1]+x[0], x[1]-x[0]])
        rects = np.repeat(dx[None, :]/2.0, len(y), axis=0) + x[:, None]

        # Make a matrix of 5-arrays (first 3 are tri, latter 2 are rect)
        trirect_mat = np.concatenate([np.repeat(triy[:, None, :], len(rects),
                                                axis=1),
                                      np.repeat(rects[None, :, :], len(triy),
                                                axis=0)],
                                     axis=2)

        # Now compute the contribution for each pair of triangle+rectangle

        with np.errstate(divide='ignore'):
            sl1 = 2.0/((trirect_mat[:, :, 2]-trirect_mat[:, :, 0]) *
                       (trirect_mat[:, :, 1]-trirect_mat[:, :, 0]))
            sl2 = 2.0/((trirect_mat[:, :, 2]-trirect_mat[:, :, 0]) *
                       (trirect_mat[:, :, 2]-trirect_mat[:, :, 1]))

        # Fix NaNs and Infs
        sl1 = np.where(np.isnan(sl1) | np.isinf(sl1), 0, sl1)
        sl2 = np.where(np.isnan(sl2) | np.isinf(sl2), 0, sl2)

        f123 = np.clip(trirect_mat[:, :, :3], trirect_mat[:, :, 3, None],
                       trirect_mat[:, :, 4, None])

        ymat = np.where((trirect_mat[:, :, 0] > trirect_mat[:, :, 4]) |
                        (trirect_mat[:, :, 2] < trirect_mat[:, :, 3]),
                        0.0,
                        ((f123[:, :, 1]-trirect_mat[:, :, 0])**2 -
                         (f123[:, :, 0]-trirect_mat[:, :, 0])**2)*0.5*sl1 +
                        ((-f123[:, :, 1]+trirect_mat[:, :, 2])**2 -
                         (-f123[:, :, 2]+trirect_mat[:, :, 2])**2)*0.5*sl2)
        ymat *= triweights[:, None]

        y = np.sum(ymat, axis=0)

        return y




# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Phylogenetic clustering class definitions"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
import warnings
import numpy as np
from scipy.cluster import hierarchy, vq
from scipy.spatial import distance as spdist
# 2-to-3 compatibility
try:
    import cPickle as pickle
except ImportError:
    import pickle
# Internal imports
from soprano.utils import get_sklearn_clusters
from soprano.collection import AtomsCollection
from soprano.analyse.phylogen.genes import (Gene, GeneDictionary,
                                            GeneError, load_genefile)
from soprano.analyse.phylogen import mapping

class PhylogenCluster(object):

    """PhylogenCluster

    An object that, given an AtomsCollection and a series of "genes" and
    weights, will build clusters out of the structures in the collection based
    on their reciprocal positions as points in a multi-dimensional space
    defined by those "genes".
    """

    def __init__(self, coll, genes=None,
                 norm_range=(0.0, 1.0),
                 norm_dist=1.0):
        """Initialize the PhylogenCluster object.

        | Args:
        |   coll (AtomsCollection): an AtomsCollection containing the
        |                           structures that should be classified.
        |                           This will be copied and frozen for the
        |                           entirety of the life of this instance;
        |                           in order to operate on a modified
        |                           collection, a new PhylogenCluster should
        |                           be created.
        |   genes (list[tuple], str, file): list of the genes that should be
        |                                   loaded immediately; each gene
        |                                   comes in the form of a tuple
        |                                   (name (str), weight (float),
        |                                   params (dict)). A path or open
        |                                   file can also be passed for a
        |                                   .gene file, from which the values
        |                                   will be loaded.
        |   norm_range (list[float?]): ranges to constrain the values of
        |                              single genes in between. Default is
        |                              (0, 1). A value of "None" in either
        |                              place can be used to indicate no
        |                              normalization on one or both sides.
        |   norm_dist (float?): value to normalize distance genes to. These
        |                       are the genes that only make sense on pairs of
        |                       structures. Their minimum value is always 0.
        |                       This number would become their maximum value,
        |                       or can be set to None to avoid normalization.

        """

        # Start by copying the collection
        if not isinstance(coll, AtomsCollection):
            raise TypeError('coll must be an AtomsCollection')
        self._collection = copy.deepcopy(coll)

        # Now initialize all the main fields
        self._genes = []
        self._gene_storage = {}  # Contains all past calculated genes
        # This one stores the norm range used last time
        self._normrng = norm_range
        self._normdist = norm_dist
        # These need to be recalculated every time the genes change
        # They also depend on stuff like normalization conditions
        self._needs_recalc = False

        self._gene_legend = None  # Legend of genes as assigned
        self._gene_vectors_raw = None   # For genes that work as points
        self._gene_matrices_raw = None  # For genes that work as distances
        self._gene_vectors_norm = None  # Normalised version of the above
        self._gene_matrices_norm = None  # Normalised version of the above
        self._distmat = None  # Total distance matrix (to feed to Scipy)

        # Now to actually add the genes
        if genes is not None:
            self.set_genes(genes=genes)

    def set_genes(self, genes, load_arrays=False):
        """Calculate, store and set a list of genes as used for clustering.

        | Args:
        |   genes (list[soprano.analyse.phylogen.Gene],
        |          file, str): a list of Genes to calculate and store. A path
        |                      or open file can also be passed for a .gene
        |                      file, from which the values will be loaded.
        |   load_arrays (bool): try loading the genes as arrays from the
        |                       collection before generating them. Warning:
        |                       if there are arrays named like genes but with
        |                       different contents this can lead to
        |                       unpredictable results.

        """

        # Check if it can be opened:
        try:
            temp = load_genefile(genes)
            genes = temp
        except TypeError:
            pass

        self._genes = []
        self._has_pairgenes = False

        for g in genes:
            self._genes.append(g)
            # First, check if it's already stored
            if g.name in self._gene_storage:
                # Is it the same?
                if self._gene_storage[g.name]['def'] == g:
                    continue
            # If it's not, it needs to be calculated
            gene_val = None
            self._has_pairgenes = self._has_pairgenes or\
                g.is_pair

            if load_arrays:
                # Check if it is present
                try:
                    gene_val = self._collection.get_array(g.name)
                except ValueError:
                    pass
            # And confirm that there are no NaNs inside
            if gene_val is None or np.any(np.isnan(gene_val)):
                gene_val = g.evaluate(self._collection)

            self._gene_storage[g.name] = {
                'def': g,
                'val': gene_val
            }

        self._needs_recalc = True

    def _recalc(self):
        """Recalculate all stored variables that depend on genes and ranges"""

        clen = self._collection.length
        self._gene_legend = [[], []]
        self._gene_vectors_raw = np.empty((clen, 0))
        self._gene_matrices_raw = np.empty((clen, clen, 0))

        g_vecs_weights = []
        g_mats_weights = []

        for g in self._genes:
            gene_val = self._gene_storage[g.name]['val']
            # Here we need to perform a reshaping
            if not g.is_pair and len(gene_val.shape) == 1:
                gene_val = gene_val.reshape((-1, 1))
            elif g.is_pair and len(gene_val.shape) == 2:
                gene_val = gene_val.reshape((gene_val.shape[0],
                                             gene_val.shape[1],
                                             1))
            # Now normalization and weights
            if not g.is_pair:
                # Append
                self._gene_vectors_raw = np.append(self._gene_vectors_raw,
                                                   gene_val,
                                                   axis=-1)
                gn = gene_val.shape[-1]
                g_vecs_weights += [g.weight/np.sqrt(gn)]*gn
                self._gene_legend[0].append((g.name, gn))
            else:
                # Append
                self._gene_matrices_raw = np.append(self._gene_matrices_raw,
                                                    gene_val,
                                                    axis=-1)
                gn = gene_val.shape[-1]
                g_mats_weights += [g.weight/np.sqrt(gn)]*gn
                self._gene_legend[1].append((g.name, gn))

        # Now calculate the normalized versions
        vnorm = self._gene_vectors_raw.copy()
        if self._normrng != (None, None):
            if self._normrng[0] is None:
                # Only maximum
                vnorm += self._normrng[1] - np.amax(vnorm, axis=0)
            elif self._normrng[1] is None:
                # Only minimum
                vnorm += self._normrng[0] - np.amin(vnorm, axis=0)
            else:
                vmin = np.amin(vnorm, axis=0)
                vmax = np.amax(vnorm, axis=0)
                vspan = vmax-vmin
                # Fix the risk of division by zero
                vspan = np.where(np.isclose(vspan, 0), np.inf, vspan)
                vnorm = (vnorm-vmin)/vspan
                vnorm *= self._normrng[1]-self._normrng[0]
                vnorm += self._normrng[0]
        self._gene_vectors_norm = vnorm*g_vecs_weights

        mnorm = self._gene_matrices_raw.copy()
        if self._normdist is not None:
            mmax = np.amax(mnorm, axis=(0, 1))
            if not np.isclose(mmax, 0):
                mnorm *= self._normdist/mmax
        self._gene_matrices_norm = mnorm*g_mats_weights

        # Distmat: start with vectors, add distances
        self._distmat = np.linalg.norm(self._gene_vectors_norm[:, None, :] -
                                       self._gene_vectors_norm[None, :, :],
                                       axis=-1)

        self._distmat = np.linalg.norm(np.append(self._distmat[:, :, None],
                                                 self._gene_matrices_norm,
                                                 axis=-1),
                                       axis=-1)

        self._needs_recalc = False

    def get_genome_vectors(self):
        """ Return the genome vectors in raw form (not normalized).
        The vectors refer to genes that allow to define a specific point for
        each structure.

        | Returns:
        |   genome_vectors (np.ndarray): a (collection.length, gene.length)
        |                                array, containing the whole extent
        |                                of the gene values for each structure
        |                                in the collection on each row
        |   genome_legend (list[tuple]): a list of tuples containing (name,
        |                                length) of the gene fragments in the
        |                                array
        """

        if self._needs_recalc:
            self._recalc()

        return self._gene_vectors_raw.copy(), self._gene_legend[0][:]

    def get_genome_matrices(self):
        """ Return the genome matrices in raw form (not normalized).
        The matrices refer to genes that only allow to define a distance
        between structures. The element at i,j represents the distance
        between said structures. The matrix is symmetric and has
        null diagonal.

        | Returns:
        |   genome_matrix (np.ndarray): a (collection.length,
        |                                collection.length, gene.length)
        |                                array, containing the distances for
        |                                each gene and pair of structures in
        |                                row and column
        |   genome_legend (list[tuple]): a list of tuples containing (name,
        |                                length) of the gene fragments in the
        |                                array
        """

        if self._needs_recalc:
            self._recalc()

        return self._gene_matrices_raw.copy(), self._gene_legend[1][:]

    def get_genome_vectors_norm(self):
        """ Return the genome vectors in normalized and weighted form.
        The vectors refer to genes that allow to define a specific point for
        each structure.

        | Returns:
        |   genome_vectors (np.ndarray): a (collection.length, gene.length)
        |                                array, containing the whole extent
        |                                of the gene values for each structure
        |                                in the collection on each row
        |   genome_legend (list[tuple]): a list of tuples containing (name,
        |                                length) of the gene fragments in the
        |                                array
        """

        if self._needs_recalc:
            self._recalc()

        return self._gene_vectors_norm.copy(), self._gene_legend[0][:]

    def get_genome_matrices_norm(self):
        """ Return the genome matrices in normalized and weighted form.
        The matrices refer to genes that only allow to define a distance
        between structures. The element at i,j represents the distance
        between said structures. The matrix is symmetric and has
        null diagonal.

        | Returns:
        |   genome_matrix (np.ndarray): a (collection.length,
        |                                collection.length, gene.length)
        |                                array, containing the distances for
        |                                each gene and pair of structures in
        |                                row and column
        |   genome_legend (list[tuple]): a list of tuples containing (name,
        |                                length) of the gene fragments in the
        |                                array
        """

        if self._needs_recalc:
            self._recalc()

        return self._gene_matrices_norm.copy(), self._gene_legend[1][:]

    def get_distmat(self):
        """Get the distance matrix between structures in the collection,
        based on the genes currently in use.

        | Returns:
        |   distmat (np.ndarray): a (collection.length, collection.length)
        |                         array, containing the overall distance
        |                         (the norm of all individual gene distances)
        |                         between all pairs of structures.

        """

        if self._needs_recalc:
            self._recalc()

        return self._distmat

    def get_linkage(self, method='single'):
        """Get the linkage matrix between structures in the collection,
        based on the genes currently in use. Only used in hierarchical
        clustering.

        Calls scipy.cluster.hierarchy.linkage.

        | Args:
        |   method (str): clustering method to employ. Valid entries are
        |                 'single', 'complete', 'weighted' and 'average'.
        |                 Refer to Scipy documentation for further details.

        | Returns:
        |   Z (np.ndarray): linkage matrix for the structures in the
        |                   collection. Refer to Scipy documentation for
        |                   details about the method

        """

        if self._needs_recalc:
            self._recalc()

        cdist = spdist.squareform(self._distmat)

        return hierarchy.linkage(cdist, method=method)

    def get_hier_clusters(self, t, method='single'):
        """Get multiple clusters (in the form of a list of collections) based
        on the hierarchical clustering methods and the currently set genes.

        Calls scipy.cluster.hierarchy.fcluster

        | Args:
        |   t (float): minimum distance of separation required to consider
        |              two clusters separate. This controls the number of
        |              clusters: a smaller value will produce more fine
        |              grained clustering. At the limit, a value smaller than
        |              the distance between the two closest structures will
        |              return a cluster for each structure. Remember that the
        |              'distances' in this case refer to distances between the
        |              'gene' values attributed to each structure. In other
        |              words they are a function of the chosen genes,
        |              normalization conditions and weights employed.
        |              In addition, the way they are calculated depends on the
        |              choice of method.
        |   method (str): clustering method to employ. Valid entries are
        |                 'single', 'complete', 'weighted' and 'average'.
        |                 Refer to Scipy documentation for further details.

        | Returns:
        |   clusters (tuple(list[int],
        |                   list[slices])): list of cluster index for each
        |                                   structure (counting from 1) and
        |                                   list of slices defining the
        |                                   clusters as formed by hierarchical
        |                                   algorithm.

        """

        if self._needs_recalc:
            self._recalc()

        Z = self.get_linkage(method=method)
        clusts = hierarchy.fcluster(Z, t, criterion='distance')
        clust_n = np.amax(clusts)
        clust_slices = [np.where(clusts == i)[0] for i in range(1, clust_n+1)]

        return clusts, clust_slices

    def get_hier_tree(self, method='single'):
        """Get a tree data structure describing the clustering order of based
        on the hierarchical clustering methods and the currently set genes.

        Calls scipy.cluster.hierarchy.to_tree

        | Args:
        |   method (str): clustering method to employ. Valid entries are
        |                 'single', 'complete', 'weighted' and 'average'.
        |                 Refer to Scipy documentation for further details.

        | Returns:
        |   root_node (ClusterNode): the root node of the tree. Access child
        |                            members with .left and .right, while .id
        |                            holds the number of the corresponding
        |                            cluster. Refer to Scipy documentation for
        |                            further details.

        """

        if self._needs_recalc:
            self._recalc()

        Z = self.get_linkage(method=method)
        return hierarchy.to_tree(Z)

    def get_max_cluster_dist(self):
        """Return the maximum possible distance between two clusters"""

        return np.linalg.norm([g.weight for g in self._genes])

    def get_kmeans_clusters(self, n):
        """Get a given number of clusters (in the form of a list of
        collections) based on the k-means clustering methods
        and the currently set genes.
        Warning: this method only works if there are no genes that work only
        with pairs of structures - as specific points, and not just distances
        between them, are required for this algorithm.

        Calls scipy.cluster.vq.kmeans

        | Args:
        |   n (int):    the desired number of clusters.

        | Returns:
        |   clusters (tuple(list[int],
        |                   list[slices])): list of cluster index for each
        |                                   structure (counting from 1) and
        |                                   list of slices defining the
        |                                   clusters as formed by k-means
        |                                   algorithm.


        """

        # Sanity check
        if self._has_pairgenes:
            # Then this method can't work!
            raise RuntimeError('k-means clustering can not be used with'
                               'presence of pair distance genes')

        if self._needs_recalc:
            self._recalc()

        centroids, dist = vq.kmeans(self._gene_vectors_norm, n)
        clusts, cdists = vq.vq(self._gene_vectors_norm, centroids)
        clust_n = np.amax(clusts)+1
        clust_slices = [np.where(clusts == i)[0] for i in range(clust_n)]
        clusts += 1

        return clusts, clust_slices

    def get_sklearn_clusters(self, method, params={}):
        """Get clusters applying any of the methods provided by the library
        scikit-learn (requires a separate installation).
        Warning: this method only works if there are no genes that work only
        with pairs of structures - as use of pairwise clustering methods is
        not implemented yet.

        Uses the sklearn.cluster.<method> class

        | Args:
        |   method (str): name of the clustering class from sklearn.clusters
        |                 to use. For reference check the documentation at
        |                 http://scikit-learn.org/stable/modules/clustering.html
        |   params (dict): parameters to be passed to the class when
        |                  initialising it. Change depending on the desired
        |                  method. Check the documentation for the specific
        |                  class.

        | Returns:
        |   clusters (tuple(list[int],
        |                   list[slices])): list of cluster index for each
        |                                   structure (counting from 1) and
        |                                   list of slices defining the
        |                                   clusters as formed by the
        |                                   requested algorithm.

        """

        # Sanity check
        if self._has_pairgenes:
            # Then this method can't work!
            raise RuntimeError('k-means clustering can not be used with'
                               'presence of pair distance genes')

        if self._needs_recalc:
            self._recalc()

        clusts = get_sklearn_clusters(self._gene_vectors_norm, method, params)
        clust_n = np.amax(clusts)+1
        clust_slices = [np.where(clusts == i)[0] for i in range(clust_n)]
        clusts += 1

        return clusts, clust_slices

    def get_clusters(self, method, params={}):
        """Wrapper method to get clusters by any available method. Depending
        on the value passed as 'method' it calls either ger_hier_clusters,
        get_kmeans_clusters, or get_sklearn_clusters. Check their respective
        docstrings for more detailed info.

        | Args:
        |   method (str): name of the clustering method to use. Can be 'hier',
        |                 'kmeans', or one of the methods in sklearn.clusters.
        |   params (dict): parameters to be passed to the class when
        |                  initialising it. Change depending on the desired
        |                  method. Check the documentation for the specific
        |                  class.

        | Returns:
        |   clusters (tuple(list[int],
        |                   list[slices])): list of cluster index for each
        |                                   structure (counting from 1) and
        |                                   list of slices defining the
        |                                   clusters as formed by the
        |                                   requested algorithm.

        """

        if method == 'hier':
            try:
                return self.get_hier_clusters(params['t'])
            except KeyError:
                raise ValueError('Parameter t required for hierarchical'
                                 ' clustering')
        elif method == 'kmeans':
            try:
                return self.get_kmeans_clusters(params['n'])
            except KeyError:
                raise ValueError('Parameter n required for k-means'
                                 ' clustering')
        else:
            return self.get_sklearn_clusters(method, params)

    def get_cluster_stats(self, clusters, raw=False):
        """Compute average values and standard deviation for each gene within
        a given clustering.

        | Args:
        |   clusters (tuple): the clustering in tuple form, as returned by one
        |                     of the get_clusters methods.
        |   raw (bool): if True, return average and standard deviation of raw
        |               instead of normalised gene values. Default is False.

        | Returns:
        |   avgs (np.ndarray): 2D array of average values of each gene for 
        |                      each cluster.
        |   stds (np.ndarray): 2D array of standard deviations of each gene 
        |                      for each cluster.
        |   genome_legend (list[tuple]): a list of tuples containing (name,
        |                                length) of the gene fragments in the
        |                                arrays
        """

        # Sanity check
        if self._has_pairgenes:
            warnings.warn('Pair distance gene stats can not be calculated')

        if raw:
            gv = self._gene_vectors_raw.copy()
        else:
            gv = self._gene_vectors_norm.copy()

        inds, slices = clusters

        avgs = np.zeros((len(slices), gv.shape[1]))
        stds = np.zeros((len(slices), gv.shape[1]))

        for i, sl in enumerate(slices):
            avgs[i] = np.average(gv[sl], axis=0)
            stds[i] = np.std(gv[sl], axis=0)

        return avgs, stds, self._gene_legend[0][:]

    def get_elbow_plot(self, method='kmeans', param_name='n',
                       param_range=range(1, 11)):
        """Returns data for an elbow plot by scanning the outcome of a given
        clustering method within a range of values for a chosen parameter.
        Used to determine optimal parameter values.

        | Args:
        |   method (str): name of the clustering method to use. Can be 'hier',
        |                 'kmeans', or one of the methods in sklearn.clusters.
        |                 Default is kmeans.
        |   param_name (str): parameter to be scanned over. Change depending
        |                     on the desired method. Check the documentation
        |                     for the specific class. Default is n, number of
        |                     clusters for k-means method.
        |   param_range (list): values of param_name to scan over. Default is
        |                       the integers from 1 to 10.

        | Returns:
        |   wss (np.ndarray): values of the "Within cluster Sum of Squares"
        |                     (WSS) to be used on the elbow plot y axis.
        |   param_range (list): range used for parameter scan, to be used on
        |                       the x axis (same as passed by the user).
        """

        wss = []

        for pval in param_range:
            clusts = self.get_clusters(method, params={param_name: pval})
            stats = self.get_cluster_stats(clusts)
            wss.append(np.sum(stats[1]**2))

        return np.array(wss), param_range

    def create_mapping(self, method="total-principal"):
        """Return an array of 2-dimensional points representing a reduced
        dimensionality mapping of the given genes using the algorithm of
        choice. All algorithms are described in [W. Siedlecki et al., Patt.
        Recog. vol. 21, num. 5, pp. 411 429 (1988)].

        | Args:
        |   method (str): can be one of the following algorithms:
        |                     - total_principal (default)
        |                     - clafic
        |                     - fukunaga-koontz
        |                     - optimal-discriminant

        """

        # Sanity check
        if self._has_pairgenes:
            # Then this method can't work!
            raise RuntimeError('No mapping can be performed in presence of'
                               'pair distance genes')

        if self._needs_recalc:
            self._recalc()

        algos = {
            'total-principal': mapping.total_principal_component,
            'clafic': mapping.classcond_principal_component,
            'fukunaga-koontz': mapping.standard_classcond_component,
            'optimal-discriminant': mapping.optimal_discriminant_plane
        }

        try:
            return algos[method](self._gene_vectors_norm)
        except KeyError:
            raise ValueError(('Invalid method passed to create_mapping.\n'
                              'Valid methods are: \n-') +
                             ('\n-'.join(algos.keys())))

    def save_collection(self, filename):
        """Save as pickle the collection bound to this PhylogenCluster.
        The calculated genes are also stored in it as arrays for future use.

        """

        saveC = self._collection[:]

        for g in self._gene_storage:
            saveC.set_array(g, self._gene_storage[g]['val'])

        saveC.save(filename)

    def save(self, filename):
        """Simply save a pickled copy to a given file path"""

        f = open(filename, 'w')
        pickle.dump(self, f)

    @staticmethod
    def load(filename):
        """Load a pickled copy from a given file path"""

        f = open(filename)
        f = pickle.load(f)
        if not isinstance(f, PhylogenCluster):
            raise ValueError('File does not contain a PhylogenCluster object')
        return f



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


""" A script that allows to run a phylogenetic analysis as a command-line
tool"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import json
import numpy as np
import argparse as ap
from soprano.collection import AtomsCollection
from soprano.properties.basic import CalcEnergy
from soprano.analyse.phylogen import PhylogenCluster, load_genefile

try:
    import matplotlib.pyplot as plt
    _has_matplotlib = True
except:
    _has_matplotlib = False


def genomes2png(aC, pC, outf, dpi=90):
    """Produce a PNG plot of a series of genomes vectors"""

    genomes, legend = pC.get_genome_vectors_norm()
    h, w = genomes.shape
    # Now build the labels
    xlabels = ['{0}_{1}'.format(l[0], i+1)
               for l in legend
               for i in range(l[1])]
    ylabels = aC.all.map(lambda s: s.info['name'])
    # Calculate the size
    figw = (w*90.0+200.0)
    figh = (h*50.0+100.0)
    plt.figure(figsize=(figw/dpi, figh/dpi), dpi=dpi)
    plt.subplots_adjust(left=150.0/figw, right=1.0-50.0/figw, bottom=50.0/figh,
                        top=1.0-50.0/figh)
    plt.pcolor(genomes, edgecolors='black', cmap='jet')
    plt.tick_params(labelsize=10)
    plt.xticks(np.linspace(0, w-1, w)+0.5, xlabels)
    plt.yticks(np.linspace(0, h-1, h)+0.5, ylabels)
    plt.savefig(outf, format='png', dpi=dpi)


def tree2json(aC, pC, w=500, h=500):
    """Produce a dictionary representation of a Scipy cluster tree object
    suitable to be saved as JSON"""

    max_dist = pC.get_max_cluster_dist()

    # First find absolute minimum and maximum energy (for hue)

    struct_Es = CalcEnergy.get(aC)

    Emin = min(struct_Es)
    Emax = max(struct_Es)

    # Create the master node by using the tree root, then visit recursively

    def visit_tree(node):

        datastr = {'value': node.get_count(),
                   'id': node.get_id(),
                   'children': []}

        # Add label text if possible

        for br in (node.left, node.right):
            if br is None:
                continue
            datastr['children'].append(visit_tree(br))

        # Now the hue and the struct name. If this is a leaf, just use the
        # structure value
        if len(datastr['children']) == 0:
            if Emax > Emin:
                hue = (Emax - CalcEnergy.get(aC.structures[node.get_id()])) \
                    / (Emax - Emin)*240.0
            else:
                hue = 120.0
            struct_name = aC.structures[node.get_id()].info['name']
        else:
            # Otherwise, an average
            hue = 0.0
            struct_name = ""
            for c in datastr['children']:
                hue += c['hue']
                struct_name += c['struct_name'] + ',\n'
            hue /= len(datastr['children'])
            struct_name = struct_name[:-2]  # Remove last newline and comma

        datastr['struct_name'] = struct_name
        datastr['hue'] = hue

        return datastr

    return visit_tree(pC.get_hier_tree())

# Entry point
def __main__():
    # So, first parse the command line arguments
    parser = ap.ArgumentParser()
    # Main argument
    parser.add_argument('seedname', type=str, default=None,
                        help="Seedname of the job")
    parser.add_argument('input_files', type=str, nargs='+', default=None,
                        help="Files to load and analyse")
    # Optional arguments
    # Boolean args
    parser.add_argument('-pkl',   action='store_true', default=False,
                        help="Input is one or more AtomsCollections"
                        " in PICKLE format")
    parser.add_argument('-tree',  action='store_true', default=False,
                        help="Create JSON tree file")
    parser.add_argument('-distmat',  action='store_true', default=False,
                        help="Save a distance matrix file")
    parser.add_argument('-forcemat',   action='store_true', default=False,
                        help="Create JSON distance matrix file for force "
                        "layout")
    parser.add_argument('-genomes',   action='store_true', default=False,
                        help="Save genomes text file")
    parser.add_argument('-gene_png',   action='store_true', default=False,
                        help="Save genomes png file")
    parser.add_argument('-gene_clamp',   action='store_true', default=True,
                        help="Clamp genes between 0 and 1")
    # Value args
    parser.add_argument('-n', type=int, default=None,
                        help="Number of lowest energy structures to keep")
    # parser.add_argument('-Z', type=int, default=None,
    # help="Expected numbers of molecules in cell (discard structures "
    # "that don't agree)")
    parser.add_argument('-corrmat_minforce',     type=float, default=0.01,
                        help="Correlation matrix minimum force")
    parser.add_argument('-corrmat_maxforce',     type=float, default=1.0,
                        help="Correlation matrix maximum force")
    parser.add_argument('-savepkl',  type=str, default=None,
                        help="Directory to save collection as pickle file")

    args = parser.parse_args()

    # First, parse the gene list
    gene_list = load_genefile(args.seedname + '.gene')

    # Then create an AtomsCollection
    if not args.pkl:
        aC = AtomsCollection(args.input_files,
                             info={'name': args.seedname},
                             cell_reduce=True,
                             progress=True)
    else:
        aC = AtomsCollection([])
        for f in args.input_files:
            aC += AtomsCollection.load(f)

    # Take only n lowest energies if required
    if (args.n is not None):
        # Sort by energy
        CalcEnergy.get(aC, store_array=True)
        aC = aC.sorted_byarray('calc_energy')[:args.n]

    #discard_Z = args.Z is not None

    # Calculate genomes
    nrange = (0, 1) if args.gene_clamp else (None, None)
    ndist = 1.0 if args.gene_clamp else None
    pC = PhylogenCluster(aC, genes=[], norm_range=nrange, norm_dist=ndist)
    pC.set_genes(gene_list, load_arrays=True)

    if args.savepkl is not None:
        # Save the structure
        pC.save_collection(os.path.join(args.savepkl, args.seedname + '.pkl'))

    # And now for the output
    if args.genomes:
        with open(args.seedname + "_genomes.dat", 'w') as ofile:
            genvec, legend = pC.get_genome_vectors()
            # Create a header
            header = ('Structure\t'
                      '{0}\n').format('\t'.join(['\t'.join(
                          ['{0}_{1}'.format(l[0],
                                            i+1)
                           for i in range(l[1])])
                          for l in legend]))
            ofile.write(header)
            try:
                for i, gen in enumerate(genvec):
                    ofile.write(aC.structures[i].info['name'] +
                                '\t' + ' '.join([str(x) for x in gen]) + '\n')
            except KeyError:
                # Someone has no name!
                raise RuntimeError('Not all structures have a name, '
                                   'can\'t print out genomes as table')

    if args.gene_png and _has_matplotlib:
        with open(args.seedname + "_genomes.png", 'w') as ofile:
            genomes2png(aC, pC, ofile)

    if args.tree:
        tree = pC.get_hier_tree()
        with open(args.seedname + "_tree.json", 'w') as ofile:
            json.dump(tree2json(aC, pC), ofile)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Definitions for the various genes used by PhylogenCluster"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import itertools
import numpy as np
from soprano.selection import AtomSelection
from soprano.utils import list_distance, is_string, compute_asymmetric_distmat
from soprano.properties.basic import LatticeCart, LatticeABC, CalcEnergy
from soprano.properties.linkage import (LinkageList, MoleculeNumber,
                                        MoleculeMass, MoleculeCOMLinkage,
                                        MoleculeRelativeRotation,
                                        HydrogenBonds, HydrogenBondsNumber,
                                        CoordinationHistogram)
from soprano.properties.labeling import (MoleculeSites, HydrogenBondTypes)
from soprano.properties.order import BondOrder


# Useful functions for parsing of complex genes
def _int_array(size=0):
    def parser(s):
        try:
            return np.array([int(x) for x in re.split('[\s,]+', s, size)])
        except:
            raise RuntimeError(('Could not parse line {0}'
                                ' as [int]*{1} array').format(s, size))


def _float_array(size=0):
    def parser(s):
        try:
            return np.array([float(x) for x in re.split('[\s,]+', s, size)])
        except:
            raise RuntimeError(('Could not parse line {0}'
                                ' as [float]*{1} array').format(s, size))


class Gene(object):

    """Gene

    A description of a property, a 'gene', characterizing a structure, to be
    used with a PhylogenCluster. A number of default genes is provided, but
    custom ones can be created as well by passing a parser. Only default genes
    can be used in a .genefile with the phylogen.py script though.

    | Args:
    |   name (str): name of the gene. Must be one of the existing ones or a
    |               custom one (in which case a parser must be provided as
    |               well). Custom names can't conflict with existing ones
    |   weight (float): weight of the gene to be applied, default is 1.0
    |   params (dict): additional parameters to be passed to the gene parser
    |                  function; when not specified, defaults will be used
    |   parser (function<AtomsCollection, \*\*kwargs>
    |           => np.array): parser function to be used when defining custom
    |                         genes. Must return a two-dimensional Numpy array
    |                         (axis 0 for the elements of the collection,
    |                          axis 1 for the values of the gene)
    |   is_pair (bool): False if the gene returns a multi dimensional point
    |                   for each structure, True if it only returns pair
    |                   distances. Default is False

    """

    def __init__(self, name, weight=1.0, params={}, parser=None, pair=False):

        # Is the gene default or custom?
        try:
            gdef = GeneDictionary.get_gene(name)
            if parser is not None:
                raise ValueError('A default gene of name {0} already exists'
                                 .format(name))
            self._parser = gdef['parser']
            self._pair = gdef['pair']
            # Check the validity of parameters as well
            if any([p not in gdef['default_params'] for p in params]):
                raise ValueError('Invalid parameters passed for gene {0}'
                                 .format(name))
        except KeyError:
            # Custom!
            if parser is None:
                raise RuntimeError('A parser function is required to create'
                                   ' custom gene of name {0}'.format(name))
            self._parser = parser
            self._pair = pair

        self.name = name
        self.weight = weight
        self.params = params

    def __eq__(self, other):

        try:
            return (self.name == other.name) and\
                   (self.weight == other.weight) and\
                   (self.params == other.params)
        except:
            return False

    @property
    def is_pair(self):
        """Whether the gene can only compare a pair of structures or can also
        give an absolute value for each structure individually (required for
        k-means clustering)"""
        return self._pair

    def evaluate(self, c):
        """Evaluate the gene on a given AtomsCollection"""
        val = self._parser(c, **self.params)

        # Check for various possible modes of failure
        if val is None or None in val:
            raise GeneError('Gene {0} has some or all None values'.format(self.name))

        try:
            if np.any(np.isnan(val)):
                raise GeneError('Gene {0} has some or all nan values'.format(self.name))
        except TypeError:
            raise GeneError('Gene {0} has values of a non-numeric type'.format(self.name))

        return val


class GeneError(Exception):
    pass


"""
Gene definitions and parser functions.

The default interface of a gene parser is:

def generic_parser(AtomsCollection, **parameters):
    return np.array [shape = (AtomsCollection.length, :)]

"""


def parsegene_energy(c):
    return np.array(CalcEnergy.get(c))


def parsegene_latt_cart(c):
    lattCartProp = LatticeCart(shape=(9,))
    return np.array(lattCartProp(c))


def parsegene_latt_abc(c):
    return np.array(LatticeABC.get(c))[:, 0]


def parsegene_latt_ang(c):
    return np.array(LatticeABC.get(c))[:, 1]


def parsegene_linkage_list(c, size=0):
    linkl = LinkageList(size=size)
    return np.array(linkl(c))


def parsegene_mol_num(c):
    return np.array(MoleculeNumber.get(c))[:, None]


def parsegene_mol_m(c, Z=0):
    molm = MoleculeMass(size=Z)
    return np.array(molm(c))


def parsegene_mol_com(c, Z=0):
    molc = MoleculeCOMLinkage(size=int(Z*(Z-1)/2))
    return np.array(molc(c))


def parsegene_mol_rot(c, Z=0, twist_axis=None, swing_plane=None):
    molr = MoleculeRelativeRotation(size=int(Z*(Z-1)/2),
                                    twist_axis=twist_axis,
                                    swing_plane=swing_plane)
    return np.array(molr(c))


def parsegene_hbonds_totn(c):
    hblen = HydrogenBondsNumber.get(c)
    return np.array([sum(ld.values()) for ld in hblen])


def parsegene_hbonds_fprint(c):
    hblen = HydrogenBondsNumber.get(c)
    return np.array([zip(*sorted(zip(ld.keys(), ld.values())))[1]
                     for ld in hblen])


def parsegene_hbonds_length(c):
    hblist = HydrogenBonds.get(c)
    hbnlist = HydrogenBondsNumber.get(c)
    # Grab the maximum number for each of these
    hbnmax = np.amax(np.array(hbnlist), axis=0)

    # Now actually gather the lengths
    def cap_to(a, n):
        return a + [np.inf]*(n-len(a))
    hblens = []
    for hbn in hblist:
        # Extract lengths, order by key
        hblen = [sorted(cap_to([hb['length'] for hb in hbn[hbs]],
                               hbnmax[hbs]))
                 for hbs in hbn]
        hblen = list(itertools.chain(*hblen))
        hblens.append(hblen)

    return np.array(hblens)


def parsegene_hbonds_angle(c):
    hblist = HydrogenBonds.get(c)
    hbnlist = HydrogenBondsNumber.get(c)
    # Grab the maximum number for each of these
    hbnmax = np.amax(np.array(hbnlist), axis=0)

    # Now actually gather the lengths
    def cap_to(a, n):
        return a + [np.inf]*(n-len(a))
    hblens = []
    for hbn in hblist:
        # Extract lengths, order by key
        hblen = [sorted(cap_to([hb['angle'] for hb in hbn[hbs]],
                               hbnmax[hbs]))
                 for hbs in hbn]
        hblen = list(itertools.chain(*hblen))
        hblens.append(hblen)

    return np.array(hblens)


def parsegene_hbonds_site_reference(c, ref=None):

    if ref is None:
        ref = c.structures[0]

    # First, calculate molecules, molecule sites and hydrogen bonds
    MoleculeSites.get(ref)
    HydrogenBonds.get(ref)

    MoleculeSites.get(c)
    HydrogenBonds.get(c)

    # Now for the actual comparison we need to compile a list of Hbonds
    # for each structure, expressed in terms of molecular sites

    reflabels = HydrogenBondTypes.get(ref)
    hblabels = HydrogenBondTypes.get(c)

    # And now to actually create a comparison
    distL = []

    for hb_lab in hblabels:
        distL.append(list_distance(hb_lab, reflabels))

    return np.array(distL)


def parsegene_hbonds_site_compare(c):

    # First, calculate molecules, molecule sites and hydrogen bonds
    MoleculeSites.get(c)
    HydrogenBonds.get(c)

    # Now for the actual comparison we need to compile a list of Hbonds
    # for each structure, expressed in terms of molecular sites

    hblabels = HydrogenBondTypes.get(c)

    # And now to actually create a comparison
    distM = np.zeros((c.length, c.length))

    for hb_i1, hb_lab1 in enumerate(hblabels):
        for hb_i2, hb_lab2 in enumerate(hblabels[hb_i1+1:]):
            d = list_distance(hb_lab1, hb_lab2)
            d /= (len(hb_lab1)+len(hb_lab2))*0.5
            distM[hb_i1, hb_i1+hb_i2+1] = d
            distM[hb_i1+hb_i2+1, hb_i1] = d

    return distM


def parsegene_coord_histogram(c, s1='C', s2='H', max_coord=6):

    chist = CoordinationHistogram(species_1=s1, species_2=s2,
                                  max_coord=max_coord)

    hists = np.zeros((c.length, max_coord+1))
    for i, h in enumerate(chist(c)):
        hists[i] = h[s1][s2]

    return hists.astype(int)


def parsegene_bond_order(c, s1=None, s2=None, channels=10, cutoff_radius=2.0,
                         cutoff_width=0.05, mode='Q'):

    if mode not in ('Q', 'W', 'QW'):
        raise ValueError('Invalid mode argument for parsegene_bond_order')

    l_channels = range(1, channels+1)

    b_ord = []

    def interpret_sel(s, sel):
        if sel is None:
            return AtomSelection.all(s)
        elif is_string(sel):
            return AtomSelection.from_element(s, sel)
        elif hasattr(sel, '__call__'):
            return sel(s)

    for s in c.structures:
        i1 = interpret_sel(s, s1)
        i2 = interpret_sel(s, s2)

        bo = BondOrder(l_channels=l_channels, center_atoms=i1,
                       environment_atoms=i2, cutoff_radius=cutoff_radius,
                       cutoff_width=cutoff_width, compute_W=('W' in mode))
        bo = bo(s)
        bo = np.concatenate([bo[i] for i in mode])

        b_ord.append(bo)

    return np.array(b_ord)


def parsegene_defect_asymmetric_fdist(c, index=0, struct=None):

    if struct is None:
        raise ValueError('defect_asymmetric_fdist gene requires a struct '
                         'argument')

    fp = c.all.get_scaled_positions()[:, index, :]
    return compute_asymmetric_distmat(struct, fp, linearized=False)


def parsegene_defect_asymmetric_fpos(c, index=0, struct=None):

    if struct is None:
        raise ValueError('defect_asymmetric_fdist gene requires a struct '
                         'argument')

    fp = c.all.get_scaled_positions()[:, index, :]
    _, imgs = compute_asymmetric_distmat(struct, fp,
                                         return_images=True)

    return imgs


class GeneDictionary(object):

    """Container class holding gene definitions"""

    _gene_dictionary = {
        'energy': {
            'default_params': {},
            'parser': parsegene_energy,
            'pair': False
        },
        'latt_cart': {
            'default_params': {},
            'parser': parsegene_latt_cart,
            'pair': False
        },
        'latt_abc_len': {
            'default_params': {},
            'parser': parsegene_latt_abc,
            'pair': False
        },
        'latt_abc_ang': {
            'default_params': {},
            'parser': parsegene_latt_ang,
            'pair': False
        },

        'linkage_list': {
            'default_params': {
                'size': int
            },
            'parser': parsegene_linkage_list,
            'pair': False
        },

        'molecule_number': {
            'default_params': {},
            'parser': parsegene_mol_num,
            'pair': False,
        },

        'molecule_mass': {
            'default_params': {
                'Z': int
            },
            'parser': parsegene_mol_m,
            'pair': False,
        },

        'molecule_com_linkage': {
            'default_params': {
                'Z': int
            },
            'parser': parsegene_mol_com,
            'pair': False,
        },

        'molecule_rot_linkage': {
            'default_params': {
                'Z': int,
                'twist_axis': _float_array(3),
                'swing_plane': _float_array(3),
            },
            'parser': parsegene_mol_rot,
            'pair': False,
        },

        'hbonds_tot_n': {
            'default_params': {},
            'parser': parsegene_hbonds_totn,
            'pair': False
        },

        'hbonds_fingerprint': {
            'default_params': {},
            'parser': parsegene_hbonds_fprint,
            'pair': False
        },

        'hbonds_length': {
            'default_params': {},
            'parser': parsegene_hbonds_length,
            'pair': False
        },

        'hbonds_angle': {
            'default_params': {},
            'parser': parsegene_hbonds_angle,
            'pair': False
        },

        'hbonds_site_reference': {
            'default_params': {
                'ref': None,
            },
            'parser': parsegene_hbonds_site_reference,
            'pair': False
        },

        'hbonds_site_compare': {
            'default_params': {},
            'parser': parsegene_hbonds_site_compare,
            'pair': True
        },

        'coord_histogram': {
            'default_params': {
                's1': 'C',
                's2': 'H',
                'max_coord': 6
            },
            'parser': parsegene_coord_histogram,
            'pair': False
        },

        'bond_order_pars': {
            'default_params': {
                's1': None,
                's2': None,
                'channels': 10,
                'cutoff_radius': 2.0,
                'cutoff_width': 0.05,
                'mode': 'Q'
            },
            'parser': parsegene_bond_order,
            'pair': False
        },

        'defect_asymmetric_fdist': {
            'default_params': {
                'index': 0,
                'struct': None,
            },
            'parser': parsegene_defect_asymmetric_fdist,
            'pair': True
        },

        'defect_asymmetric_fpos': {
            'default_params': {
                'index': 0,
                'struct': None,
            },
            'parser': parsegene_defect_asymmetric_fpos,
            'pair': False
        }

    }

    _gene_help = {

        'energy': """
        The calculated total energy of a structure. This is the value returned
        by ase.Atoms.get_potential_energy().

        Parameters: None
        Length: 1
        """,

        'latt_cart': """
        The lattice parameters of a structure in Cartesian form, inline.

        Parameters: None
        Length: 9
        """,

        'latt_abc_len': """
        The length of the A, B and C lattice parameters.

        Parameters: None
        Length: 3
        """,

        'latt_abc_ang': """
        The angles alpha, beta and gamma defining the lattice.

        Parameters: None
        Length: 3
        """,

        'linkage_list': """
        A list of the n shortest interatomic distances in the structure
        (periodic boundaries are taken into account).

        Parameters:
            size (int): how many distances are used (default = 10)
        Length: [size]
        """,

        'molecule_number': """
        Number of molecules found in the structure.

        Parameters: None
        Length: 1
        """,

        'molecule_mass': """
        Masses of each of the molecules found in the structure.

        Parameters:
            Z (int): expected number of molecules (default = total number of
                     molecules)
        Length: Z
        """,

        'molecule_com_linkage': """
        A list of the n shortest intermolecular distances in the structure
        (periodic boundaries are taken into account and the molecules are
         considered coincident with their Center Of Mass).

        Parameters:
            Z (int): expected number of molecules (default = total number of
                     molecules)
        Length: Z*(Z-1)/2
        """,

        'molecule_rot_linkage': """
        A list of the n shortest intermolecular rotational distances in the
        structure (orientation is considered to be the one of the molecules'
        principal inertia axis system, standard 1-|q1.q2| quaternion distance
        is used).

        Parameters:
            Z (int): expected number of molecules (default = total number of
                     molecules)
            twist_axis ([float]*3): if present, only compare the Twist
                                    component of quaternion along the given
                                    axis. The Twist/Swing decomposition splits
                                    a quaternion in a rotation around an axis
                                    and one around an orthogonal direction.
                                    Only one between this and swing_plane can
                                    be present.
            swing_plane ([float]*3): if present, only compare the Swing
                                     component of quaternion along the given
                                     axis. The Twist/Swing decomposition
                                     splits a quaternion in a rotation around
                                     an axis and one around an orthogonal
                                     direction. Only one between this and
                                     twist_axis can be present.
        Length: Z*(Z-1)/2
        """,

        'hbonds_tot_n': """
        Total number of hydrogen bonds in the system. If the hydrogen bonds
        have already been calculated, the stored data will be used.

        Parameters: None
        Length: 1
        """,

        'hbonds_fingerprint': """
        Hydrogen bonds fingerprint, i.e. number of hydrogen bonds split by
        type. By convention the ordering is alphabetical, with the types
        described as AH..B (A, B replaced by the respective chemical symbols).
        If the hydrogen bonds have already been calculated, the stored data
        will be used.

        Parameters: None
        Length: 1
        """,

        'hbonds_length': """
        Hydrogen bonds lengths, i.e. length of all hydrogen bonds in the
        system, sorted by type and value.

        Parameters: None
        Length: sum of maximum number of hydrogen bonds per type
        """,

        'hbonds_angle': """
        Hydrogen bonds angles, i.e. amplitude of all hydrogen bonds angles in
        the system, sorted by type and value.

        Parameters: None
        Length: sum of maximum number of hydrogen bonds per type
        """,

        'hbonds_site_reference': """
        Comparison between hydrogen bonds in the given systems and a reference
        system, site-wise.
        Works for molecular systems. Bonds are compared to see how many share
        the same structure (namely, are established between equivalent sites
        in equivalent molecules). The number of different bonds is given.

        Parameters:
            ref (ase.Atoms): reference system to compare against (default =
                             first system in the collection)
        Length: 1
        """,

        'hbonds_site_compare': """
        Comparison between hydrogen bonds in two given systems, site-wise.
        Works for molecular systems. Bonds are compared to see how many share
        the same structure (namely, are established between equivalent sites
        in equivalent molecules). The number of different bonds is treated as
        a 'distance' between pairs of systems.

        Parameters: None
        Length: 1
        """,

        'coord_histogram': """
        Coordination histogram for two given species s1 and s2. Gives an array
        whose element of index i contains the number of atoms of species s1
        that are bonded to i atoms of species s2. The parameter max_coord
        controls how many bins are created; any coordination number higher
        than max_coord gets bunched up in the last element. By default this
        means all coordination values of 6 or more.
        As an example, the s1=C, s2=H histogram for a CH3CH2CH=CHCOOH molecule
        would look like: [1,2,1,1,0,0,0].

        Parameters:
            s1 (string): chemical symbol of species whose coordination is to
                         be evaluated. By default C
            s2 (string): chemical symbol of ligand species with s1. By default
                         H
            max_coord (int): maximum coordination number to allow in
                             histogram. By default 6
        Length: max_coord+1 (by default 7)
        """,

        'bond_order_pars': """
        Steinhardt bond order parameters (based on spherical harmonics) for a
        pair of species in the given system. These parameters reflect the
        symmetry of the local environment for each atom by making use of 
        spherical harmonics. Spherical harmonic functions are computed up to 
        the requested angular momentum channel; higher angular momenta
        correspond to more rapidly varying angular functions (and therefore
        will match finer details). There are two types of parameter: 
        Q (second order) and W (third order). Both are invariant to rotation,
        but the W are more expensive to compute and are off by default. A 
        sigmoidal cutoff is applied to weight atoms in the neighbourhood;
        parameters for the sigmoidal function can be passed.

        Parameters:
            s1 (multiple types): selector for the atoms whose environment is
                                 to be evaluated.
                                 If an AtomSelection is used, it will define
                                 the atoms.
                                 If a string is used, it will be
                                 interpreted as a symbol of a chemical
                                 species.
                                 If a callable is used, it must take an 
                                 ase.Atoms object as the only argument, and
                                 return an AtomSelection. 
                                 Default is None, meaning a sum over all
                                 atoms.
            s2 (multiple types): selector for the atoms that contribute to the
                                 environment.
                                 Follows the same rules as s1.
                                 Default is None, meaning a sum over all 
                                 atoms.
            channels (int): number of angular momentum channels. More channels
                            mean finer detail but higher computational cost.
                            Default is 10.
            cutoff_radius (float): cutoff distance for sigmoidal function.
                                   Default is 2 Angstroms.
            cutoff_width (float): scale over which the sigmoidal function
                                  should vanish. Default is 0.05 Angstroms.
            mode (str): mode to compute. Can be Q (only second order), W (only
                        third order) and QW (both orders, concatenated).
                        Default is Q.

        """,

        'defect_asymmetric_fdist': """
        Compute a fractional coordinates distance between single atoms in the
        structures accounting for all effects of symmetry operations. This
        should group together atoms occupying sites that are
        crystallographically equivalent. This is especially useful for defect
        analysis. A pure structure must be passed as an argument to compute
        the symmetry operations in the first place. An installation of spglib
        is required for this computation.

        Parameters:
            index (int): index of the defect for which the distance matrix is
                         to be computed. It must be the same for all
                         structures. Default is 0
            struct (ase.Atoms): pure structure from which the space group and
                                symmetry operations must be computed. Default
                                is None, must be provided for the calculation
                                to work
        """,

        'defect_asymmetric_fpos': """
        Compute an asymmetric fractional coordinates position for a single atom
        in the structures accounting for all effects of symmetry operations.
        This should group together atoms occupying sites that are
        crystallographically equivalent. This is especially useful for defect
        analysis. A pure structure must be passed as an argument to compute
        the symmetry operations in the first place. An installation of spglib
        is required for this computation.

        Parameters:
            index (int): index of the defect for which the distance matrix is
                         to be computed. It must be the same for all
                         structures. Default is 0
            struct (ase.Atoms): pure structure from which the space group and
                                symmetry operations must be computed. Default
                                is None, must be provided for the calculation
                                to work

        """
    }

    @classmethod
    def get_gene(self, g):
        """Get the definition for a given gene"""
        return dict(self._gene_dictionary[g])

    @classmethod
    def help(self, g=None):
        """Get an help string for a given gene"""
        if g is not None:
            return self._gene_help[g]
        else:
            return ('List of available genes:\n' +
                    '\n'.join(self._gene_help.keys()))


def load_genefile(gfile):
    """Load a gene file and return the (validated) list of genes contained
    within.

    | Args:
    |   gfile (file or str): file to parse

    | Returns:
    |   genelist: a list of genes parsed from the given file, ready to be
    |             passed to a PhylogenCluster

    """

    if hasattr(gfile, '__read__'):
        genefile = gfile.read()
    else:
        gfile = open(gfile)
        genefile = gfile.read()
        gfile.close()

    comment_re = re.compile('#[^\n]*\n')
    gene_re = re.compile('\s*([a-zA-Z_]+)\s+([0-9\.]+)(?:\s*{([^}]+)})*')
    block_re = re.compile('\s*([a-zA-Z_]+)\s+([0-9\.]+)')

    # First, remove comments
    genefile = comment_re.sub('\n', genefile)

    # Then actually get the blocks
    gblocks = gene_re.findall(genefile)

    genes = []

    for g in gblocks:
        gname = g[0]
        try:
            gentry = GeneDictionary.get_gene(gname)
        except:
            raise GeneError('{0} is not a valid gene'.format(gname))
        try:
            gweight = float(g[1])
        except:
            raise GeneError('Badly formatted weight definition for gene'
                            ' {0}'.format(gname))
        # Does it have an argument block?
        gargs = {}
        if g[2] != '':
            for arg in block_re.findall(g[2]):
                try:
                    arg_parser = gentry['default_params'][arg[0]]
                except:
                    raise GeneError(('{0} is not a valid parameter for'
                                     ' {1}').format(arg[0], gname))
                try:
                    gargs[arg[0]] = arg_parser(arg[1])
                except:
                    raise GeneError('Invalid value for parameter '
                                    '{0}'.format(arg[0]))

        genes.append(Gene(gname, gweight, gargs))

    return genes



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" A very simple script to convert from all ASE-accepted formats to CASTEP's
.cell format"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import glob
import argparse as ap
from ase import io


def __main__():

    parser = ap.ArgumentParser()
    # Main argument
    parser.add_argument('input_folders', type=str, nargs='+', default=None,
                        help="Folders of VASP input files to convert to CELL")

    args = parser.parse_args()

    _inpfiles = ('POSCAR', 'CONTCAR')

    for inpf in args.input_folders:
        # Check that these folders have the right files
        a = None
        for ft in _inpfiles:
            try:
                a = io.read(os.path.join(inpf, ft))
            except IOError as e:
                if "Could not determine chemical symbols" in str(e):
                    # Wrong header format!
                    print("WARNING - ASE could not read elements in "
                          "POSCAR file.\n"
                          "By convention, elements should be elencated in the "
                          "first line of the POSCAR file in the order "
                          "in which they appear in the line in which the "
                          "number of each atom type is given.")
                continue
        if a is None:
            print("No valid structure files found in folder "
                  "{0}\nSkipping...".format(inpf))
            continue

        name = os.path.split(inpf)[-1]
        io.write(name + '.cell', a)


# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Bindings for AIRSS Buildcell program for random structure generation"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import copy
import hashlib
import subprocess as sp
from ase import io as ase_io
# Internal imports
from soprano.utils import seedname, safe_communicate

# Python 2-to-3 compatibility
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO


def airssGen(input_file,
             n=100,
             buildcell_command='buildcell',
             buildcell_path=None,
             clone_calc=True):
    """Generator function binding to AIRSS' Buildcell.

    This function searches for a buildcell executable and uses it to
    generate multiple new Atoms structures for a collection.

    | Args:
    |   input_file (str or file): the .cell file with appropriate comments
    |                             specifying the details of buildcell's
    |                             construction work.
    |   n (int): number of structures to generate. If set to None the
    |            generator goes on indefinitely.
    |   buildcell_command (str): command required to call the buildcell
    |                            executable.
    |   buildcell_path (str): path where the buildcell executable can be
    |                         found. If not present, the buildcell command
    |                         will be invoked directly (assuming the
    |                         executable is in the system PATH).
    |   clone_calc (bool): if True, the CASTEP calculator in the input file
    |                      will be copied and attached to the new structures.
    |                      This means that for example any additional CASTEP
    |                      keywords/blocks in the input file will be carried
    |                      on to the new structures. Default is True.

    | Returns:
    |   airssGenerator (generator): an iterable object that yields structures
    |                               created by buildcell.

    """

    # First: check that AIRSS is even installed
    if buildcell_path is None:
        buildcell_path = ''
    airss_cmd = [os.path.join(buildcell_path, buildcell_command)]

    try:
        stdout, stderr = sp.Popen(airss_cmd + ['-h'],
                                  stdout=sp.PIPE,
                                  stderr=sp.PIPE).communicate()
    except OSError:
        # Not even installed!
        raise RuntimeError('No instance of Buildcell found on this system')

    # Now open the given input file
    try:
        input_file = open(input_file)   # If it's a string
    except TypeError:
        pass                            # If it's already a file
    template = input_file.read()
    # Now get the file name
    basename = seedname(input_file.name)
    input_file.close()

    # Calculator (if needed)
    calc = None
    if clone_calc:
        calc = ase_io.read(input_file.name).calc

    # And keep track of the count!
    # (at least if it's not infinite)
    i = 0

    while True:
        if n is not None:
            if i >= n:
                return
            i += 1

        # Generate a structure
        subproc = sp.Popen(airss_cmd,
                           universal_newlines=True,
                           stdin=sp.PIPE,
                           stdout=sp.PIPE,
                           stderr=sp.PIPE)
        stdout, stderr = safe_communicate(subproc, template)

        # Now turn it into a proper Atoms object
        # To do this we need to make it look like a file to ASE's io.read
        try:
            newcell = ase_io.read(StringIO(stdout), format='castep-cell')
        except:
            # If ANYTHING happens, let's consider that stdout might be wrong
            raise RuntimeError(('Invalid output from buildcell:\nstdout:\n{0}'
                                '\nstderr:\n{1}')
                               .format(stdout, stderr))
        if clone_calc:
            newcell.calc = copy.deepcopy(calc)
        # Generate it a name, function of its properties
        postfix = hashlib.md5(str(newcell.get_positions()
                                  ).encode()).hexdigest()
        newcell.info['name'] = '{0}_{1}'.format(basename, postfix)
        yield newcell



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" A script to compute averages of NMR tensors over CH3 and NH3 groups"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


import numpy as np
import argparse as ap

from ase import io
from soprano.properties.linkage import Bonds


def __main__():

    parser = ap.ArgumentParser(description="""
        Processes .magres files containing chemical groups of the form XH3
        (by default, CH3 and NH3), to average the required NMR tensors for the
        hydrogen atoms connected to such groups.
        """)
    # Main argument
    parser.add_argument('input_files', type=str, nargs='+', default=None,
                        help="Magres files on which to carry out averages")
    # Optional arguments
    parser.add_argument('-X', type=str, nargs='+', default=['C', 'N'],
                        help="Nuclei to consider for XH3 groups "
                        "(default: C and N)")
    parser.add_argument('-vdws', type=float, default=1.0,
                        help="Van der Waals radius scale for bond calculation."
                        " Increase for higher tolerance")
    parser.add_argument('-avg', type=str, nargs='+', default=['ms', 'efg'],
                        help="Arrays to average over XH3 groups "
                        "(default: ms and efg)")
    parser.add_argument('-prefix', type=str, default='avg',
                        help="Prefix added to output files")

    args = parser.parse_args()

    bcalc = Bonds(vdw_scale=args.vdws, return_matrix=True)

    for f in args.input_files:
        try:
            a = io.read(f)
        except IOError:
            print('File {0} not found, skipping'.format(f))
            continue

        # Do they have magres data?
        if not any([a.has(k) for k in args.avg]):
            print('File {0} has no data to average, skipping'.format(f))
            continue

        # Find what to average
        bonds, bmat = bcalc(a)

        # Find XH3 groups
        symbs = np.array(a.get_chemical_symbols())
        hinds = np.where(symbs == 'H')[0]
        h3groups = []

        for xsymb in args.X:

            xinds = np.where(symbs == xsymb)[0]
            xinds = xinds[np.where(np.sum(bmat[xinds][:, hinds],
                                          axis=1) == 3)[0]]
            if len(xinds) > 0:
                h3groups.append(np.where(bmat[xinds][:, hinds] == 1)[1])

        # Now average
        avg_a = a.copy()

        for k in args.avg:
            arr = a.get_array(k)
            for h3 in h3groups:
                arr[h3] = np.average(arr[h3], axis=0)
            avg_a.set_array(k, arr)

        io.write('{0}_{1}'.format(args.prefix, f), avg_a)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Methods used to spawn or kill multiple daemons on a cluster machine.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import shutil
from soprano.utils import import_module

# Spawn a Submitter instance from a given module
subm_module, subm_name = sys.argv[1:3]
loaded_module = import_module(subm_module)
subm = getattr(loaded_module, subm_name)
try:
    subm.start()
except Exception as e:
    # If ANYTHING goes wrong, at least clean up!
    for j in subm._jobs:
        subm.queue.kill(j)
        try:
            shutil.rmtree(subm._jobs[j]['folder'])
        except OSError:
            pass  # Whatever, it was deleted already I guess
    subm.log('Submitter crashed following error:'
             '\n{0}:\n\t{1}'.format(type(e).__name__, e))



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Utility functions for operating with GULP

"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re


def _gulp_cell_definition(s, syms=None):
    """Create a cell definition for a GULP input file. Will use syms if
    passed, otherwise standard chemical symbols"""

    gcell = "vectors\n{0}\n".format('\n'.join(['\t'.join([str(x)
                                                          for x in row])
                                               for row in s.get_cell()]))
    gcell += "frac\n"
    syms = s.get_chemical_symbols() if syms is None else syms
    pos = s.get_scaled_positions()
    for i, s in enumerate(syms):
        gcell += "{0} {1} {2} {3}\n".format(s, *pos[i])

    return gcell


def _gulp_parse_energy(lines):
    """Parse energy out of a GULP output split in lines"""
    for l in lines[::-1]:
        if 'Total lattice energy       =' in l and 'eV' in l:
            return float(l.split()[4])

    return None


def _gulp_parse_charges(lines):
    """Parse charges out of a GULP output split in lines"""
    q_re = re.compile('Final charges from\s+([a-zA-Z\-]+)\s+:')
    qline_re = re.compile('([0-9]+)\s+([0-9]+)\s+([0-9\.\-]+)')
    for i, l in enumerate(lines[::-1]):
        q_type = q_re.findall(l)
        if len(q_type) == 1:
            # Found it!
            charges = {'type': q_type[0],
                       'q': [],
                       'Z': []}
            # Go forward until you find the first valid line
            in_block = False
            for l2 in lines[(len(lines)-i):]:
                parsed = qline_re.findall(l2)
                if len(parsed) == 0:
                    if in_block:
                        return charges
                    else:
                        continue
                else:
                    in_block = True
                    charges['q'].append(float(parsed[0][2]))
                    charges['Z'].append(int(parsed[0][1]))

    return None



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Convenience util used to start/stop submitter processes"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


def submitter_handler():

    import os
    import sys
    import argparse as ap
    import subprocess as sp
    from soprano.utils import import_module
    from soprano.hpc.submitter import Submitter

    # First of all, check that we are on Linux, this only works there
    if sys.platform not in ('linux', 'linux2'):
        sys.exit('Invalid OS: background Submitter launching works only on '
                 'Linux')

    parser = ap.ArgumentParser(description="Use this script to start or stop "
                               "a Submitter object as a background process.")
    # Required arguments
    parser.add_argument('action', type=str, nargs=1,
                        help="Action to perform: "
                             "start -> start the given submitter, "
                             "stop -> stop the given submitter if running, "
                             "list -> list the currently running submitters"
                             " from the given file,"
                             " or any custom action defined for your"
                             " submitter")
    parser.add_argument('submitter_file', type=str,
                        help="Name of the Python module file containing the"
                             " declaration for the Submitter to use")
    # Optional arguments
    parser.add_argument('-n', type=str, default=None,
                        help="Name of the Submitter instance to use, if "
                        "more than one is present in the given file. "
                        "CAREFUL: this is the name of the variable, not the "
                        "one passed as parameter in the constructor.")
    parser.add_argument('-nohup', action='store_true', default=False,
                        help="If True, nohup is used to make sure that the "
                        "process does not quit in case the user logs out "
                        "(for example on a login node of an HPC machine).")

    try:
        args = parser.parse_args()
    except ValueError:
        sys.exit('Invalid submitter_file argument: file is not a valid Python'
                 ' module')

    # First, check that the required submitter file exists and that it
    # contains what we need
    abs_subm_file = os.path.abspath(args.submitter_file)
    loaded_module = import_module(abs_subm_file)

    # Now load the actual Submitter!
    subms = {v: getattr(loaded_module, v) for v in dir(loaded_module)
             if v[:2] != '__' and
             issubclass(getattr(loaded_module, v).__class__, Submitter)}

    if len(subms) == 0:
        sys.exit('No Submitter instance found in '
                 '{0}'.format(args.submitter_file))
    elif len(subms) > 1 and args.n is None:
        sys.exit('Too many Submitter instances found in '
                 '{0}'.format(args.submitter_file))
    elif args.n is not None and args.n not in subms:
        sys.exit(('Submitter of name {0} '
                  'not found in {1}').format(args.n, args.submitter_file))
    # We're fine!
    if args.n is not None:
        submitter_name = args.n
    else:
        submitter_name = list(subms.keys())[0]

    if args.action[0] == 'start':
        # Ok, first let's check that nothing like that is running already
        subm_l = Submitter.list()
        for s in subm_l:
            if (os.path.abspath(s[0]) == abs_subm_file 
                and submitter_name == s[1]):
                sys.exit('The requested submitter is already'
                         ' running')
        cmd = ['python', '-m', 'soprano.hpc.submitter._spawn',
               abs_subm_file, submitter_name, '&']
        if args.nohup:
            cmd = ['nohup'] + cmd
        sp.Popen(cmd)
        print("Submitter "
              "{0} from file {1} started".format(submitter_name,
                                                 args.submitter_file))
    elif args.action[0] == 'stop':
        # PKILL the process
        succ = Submitter.stop(abs_subm_file,
                              submitter_name)
        if succ:
            print("Submitter "
                  "{0} from file {1} stopped".format(submitter_name,
                                                     args.submitter_file))
        else:
            print("The requested submitter is not running")
    elif args.action[0] == 'list':
        subm_l = Submitter.list()
        tabf = "{1: >10}\t| {2: >10}\t| {0: >10}"
        print(tabf.format('File', 'Name', 'Time'))
        for s in subm_l:
            if os.path.abspath(s[0]) != abs_subm_file:
                continue
            print(tabf.format(*s))
    elif args.action[0] in subms[submitter_name]._user_signals:
        signum, _ = subms[submitter_name]._user_signals[args.action[0]]
        subm_l = Submitter.list()
        for s in subm_l:
            if (os.path.abspath(s[0]) == abs_subm_file
                    and s[1] == submitter_name):
                os.kill(s[3], signum)
                break
    else:
        print("Unknown action {0}".format(args.action[0]))

submitter_handler()



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Definition of the Collection class.

It handles multiple Atoms ASE objects and mirrors in this sense the structure
of the Atoms object itself.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import ase
import glob
import shutil
import inspect
import warnings
import numpy as np
# 2-to-3 compatibility
try:
    import cPickle as pickle
except ImportError:
    import pickle
# More 2-to-3 compatibility, defining a wrapper function for inspect
# Internal imports
from ase import io as ase_io
from ase.build import niggli_reduce
from ase.calculators.singlepoint import SinglePointCalculator
from soprano import utils

utils.customize_warnings()


class _AllCaller(object):

    """_AllCaller class.

    A meta-object that serves the purpose of calling a function on all members
    of a list in a natural way.
    """

    def __init__(self, all_list, all_class=None):
        """Initialize the AllCaller with an 'all' list"""
        if all_class is None:
            self._class = all_list[0].__class__
        else:
            self._class = all_class
        if not all([x.__class__ is self._class for x in all_list]):
            raise ValueError('Elements of list passed to an _AllCaller'
                             ' must be of the same type.')
        self._all = all_list
        # Now get a list of all members in common between these instances
        if len(self._all) > 0:
            self._instance_attrs = set.intersection(*[set(ins.__dict__.keys())
                                                      for ins in self._all])

    def __getattr__(self, name):
        """Here's the magic of the class - when a method isn't found belonging
        to it, go looking for it in its ._all list..."""

        if hasattr(self._class, name):
            attr = getattr(self._class, name)
            # Is it a function?
            if not hasattr(attr, '__call__'):
                return np.array([getattr(x, name) for x in self._all])

            def iterfunc(*args, **kwargs):
                return np.array([getattr(x, name)(*args, **kwargs)
                                 for x in self._all])
            return iterfunc
        elif name in self._instance_attrs:
            # It's an instance attribute
            return np.array([getattr(x, name) for x in self._all])
        else:
            raise AttributeError(('Not all \'{0}\' objects have attribute'
                                  ' \'{1}\'').format(self._class.__name__,
                                                     name))

    def map(self, f, *args, **kwargs):
        """Map a function to each element of the ._all list and return the
        results."""

        # First, check the signature
        if not hasattr(f, '__call__'):
            raise TypeError('Only functions can be mapped')

        nargs, nargs_def = utils.inspect_args(f)
        if nargs < 1 + len(args) + len(kwargs):
            # Function is invalid!
            raise ValueError('Invalid function passed to map')
        return [f(x, *args, **kwargs) for x in self._all]


class AtomsCollection(object):

    """AtomsCollection object.

    An AtomsCollection represents a group of ASE Atoms objects.
    It handles them together, can perform mass operations on them, and stores
    arrays of informations related to them.
    """

    def __init__(self, structures=[],
                 info={},
                 cell_reduce=False,
                 progress=False, suppress_ase_warnings=True):
        """
        Initialize the AtomsCollection

        | Args:
        |    structures (list[str] or list[ase.Atoms]): list of file names or
        |                                               Atoms that will form
        |                                               the collection
        |    info (dict): dictionary of general information to attach
        |                 to this collection
        |    cell_reduce (bool): if True, perform a Niggli cell reduction on
        |                        all loaded structures
        |    progress (bool): visualize a progress bar for the loading process
        |    suppress_ase_warnings (bool): suppress annoying ASE warnings when
        |                                  loading files (default is True)
        """

        # Start by parsing out the structures
        self.structures = []

        if isinstance(structures, ase.Atoms):
            # Well, it's just one...
            structures = [structures]
        elif inspect.isgenerator(structures):
            # Let's unravel it
            iter_structs = structures
            structures = []
            for s in iter_structs:
                structures.append(s)

        if progress:
            sys.stdout.write("Loading collection...\n")
        s_n = len(structures)
        for s_i, struct in enumerate(structures):
            if progress:
                # Progress bar
                sys.stdout.write("\rLoading: {0}".format(utils.progbar(s_i+1,
                                                                       s_n)))
            # Is it an Atoms object?
            if type(struct) is ase.Atoms:
                self.structures.append(ase.Atoms(struct))
                # Copy all arrays
                for k in struct.arrays.keys():
                    if not self.structures[-1].has(k):
                        self.structures[-1].new_array(k, struct.get_array(k))
                if struct.calc is not None:
                    # Prevents pointless attempts at re-calculating
                    self.structures[-1].calc._old_atoms = self.structures[-1]
            # Or is it a string?
            elif utils.is_string(struct):
                with utils.silence_stdio(suppress_ase_warnings,
                                         suppress_ase_warnings):
                    self.structures.append(ase_io.read(str(struct)))
                # If there's no name, give it the filename
                if 'name' not in self.structures[-1].info:
                    self.structures[-1].info['name'] = utils.seedname(struct)
            else:
                raise TypeError('Structures must be Atoms objects or valid '
                                'file names,'
                                ' not {0}'.format(type(struct).__name__))
            if cell_reduce:
                # Here we must keep the energy if it was present
                # We do this by hand because ASE has its good reasons
                # for severing the atoms-calculator connection when changing
                # the unit cell.
                try:
                    _E = self.structures[-1].calc.results['energy']
                except (KeyError, AttributeError):
                    _E = None
                niggli_reduce(self.structures[-1])
                if _E is not None:
                    _calc = SinglePointCalculator(self.structures[-1],
                                                  energy=_E)
                    self.structures[-1].set_calculator(_calc)

        if progress:
            sys.stdout.write('\nLoaded {0} structures\n'.format(s_n))

        self._all = _AllCaller(self.structures, ase.Atoms)

        self._arrays = {}

        # Now assign the info
        if type(info) is not dict:
            raise TypeError('Info must be dict,'
                            ' not {0}'.format(type(info).__name__))
        else:
            self.info = info.copy()

    def __add__(self, other):
        """Addition of two collections brings a merging"""

        if not isinstance(other, self.__class__):
            raise TypeError('\'AtomsCollection\' does not support operator +'
                            ' with object of type'
                            ' \'{0}\''.format(type(other).__name__))

        # Create a common collection, join arrays where present, copy all info
        all_struct = list(self.structures)
        all_struct += list(other.structures)

        all_info = dict(self.info)
        for k in other.info:
            if k not in all_info:
                all_info[k] = other.info[k]
            else:
                all_info.pop(k)

        sum_struct = AtomsCollection(all_struct, all_info)

        # Now arrays
        all_arrays = {}
        for aname in self._arrays:
            # Grab the shape
            shape = self._arrays[aname].shape[1:]
            # Check if present in the other as well
            all_arr = list(self.get_array(aname))
            if aname in other._arrays:
                all_arr += list(other.get_array(aname))
            else:
                all_arr += [np.zeros(shape)*np.nan]*other.length
            sum_struct.set_array(aname, all_arr, shape=shape)

        for aname in other._arrays:
            # Can not be present in both
            if aname in self._arrays:
                continue
            # Grab the shape
            shape = other._arrays[aname].shape[1:]
            all_arr = list(other.get_array(aname))
            all_arr = [np.zeros(shape)]*self.length + all_arr
            sum_struct.set_array(aname, all_arr, shape=shape)

        return sum_struct

    def __iadd__(self, other):
        self = self + other
        return self

    def __getitem__(self, indices):
        """Allow sophisticated slicing"""

        if type(indices) is int:
            # Special case, a single element!
            indices = indices%len(self)
            indices = slice(indices, indices+1)

        try:
            struct_slice = self.structures[indices]
        except TypeError:
            indices = np.array(indices)
            if indices.dtype == bool:
                indices = np.where(indices)[0]  # Support for bool arrays
            struct_slice = [self.structures[i] for i in indices]

        sliced = AtomsCollection(struct_slice, info=self.info)

        # Now to add the arrays
        for a in self._arrays:
            sliced.set_array(a, self._arrays[a][indices])

        return sliced

    def __iter__(self):
        return self.structures.__iter__()

    def __deepcopy__(self, memodict={}):
        """Protects against problems with infinite recursion in AllCaller"""

        dcopy = AtomsCollection(self.structures, self.info)
        for arr in self._arrays:
            dcopy.set_array(arr, self._arrays[arr])

        return dcopy

    def __len__(self):
        return self.length

    @property
    def length(self):
        return len(self.structures)

    @property
    def all(self):
        return self._all

    def set_array(self, name, a, dtype=None, shape=None, args={}):
        """Add or modify an array of data related to the Atoms objects
        in this collection.

        | Args:
        |   name (str): name of the array to operate on.
        |   a (np.ndarray or function<Atoms, \*\*kwargs>
        |                    => Any): the data to assign to the array (must
        |                             be same length as the collection) or
        |                             a function that takes an Atoms object
        |                             as the first argument and returns a
        |                             value. This will be mapped over the
        |                             structures to create the array.
        |   dtype (type): type to cast the values of the array to.
        |   shape (tuple [int]): shape of each entry of the array. Will be
        |                        checked if provided.
        |   args (dict): named arguments to pass to the function provided
        |                as a. Will be ignored if an array is passed instead.

        """

        # a can be an actual array or a function that operates on each
        # separate Atoms object and returns a value

        a = np.array(a, dtype)
        if a.shape == ():
            a = a.item()
            if hasattr(a, '__call__'):
                # It's a function
                a = np.array(self.all.map(a, **args), dtype)
            else:
                # Invalid
                raise TypeError('new_array requires to pass either an array'
                                ' or a function taking an Atoms object as its'
                                ' first argument and returning a value.')

        # Now check that the shape is valid
        if shape is not None:
            if shape == (1,):
                targ_shape = (self.length,)
            else:
                targ_shape = (self.length,) + shape
            if a.shape != targ_shape:
                raise ValueError('Array of invalid shape passed to new_array')
        else:
            if a.shape[0] != self.length:
                raise ValueError('Array passed to new_array should be'
                                 ' as long as the number of structures')

        # And finally, assign
        self._arrays[name] = a

    def get_array(self, name, copy=True):
        """Get a copy of an array of given name (or a reference if copy=False)

        | Args:
        |   name (str): name of the array to retrieve.
        |   copy (bool): if the array should be copied or a reference should
        |                be returned instead.

        | Returns:
        |   array (np.ndarray): the requested array

        """

        if name not in self._arrays:
            raise ValueError('Array \'{0}\' does not exist'.format(name))
        else:
            return np.array(self._arrays[name], copy=copy)

    def has(self, name):
        """Check if array of given name exists"""

        return name in self._arrays

    def set_calculators(self, calctype, labels=None, params={}):
        """Set an ASE calculator on each structure in the collection,
        and set said calculator's parameters.

        | Args:
        |   calctype (ASE Calculator type): the type of calculator
        |                                   to instantiate.
        |   labels (Optional[list[str]]): names to use for the calculators'
        |                                 files. If not present, random
        |                                 generated names are used.
        |   params (Optional[dict]): parameters of the calculator to set.

        """

        # First, a check
        from ase.calculators.general import Calculator as gCalculator
        from ase.calculators.calculator import Calculator as cCalculator
        from ase.calculators.calculator import FileIOCalculator as ioCalculator

        if (gCalculator not in calctype.__bases__) and \
           (cCalculator not in calctype.__bases__) and \
           (ioCalculator not in calctype.__bases__):
            raise TypeError('calctype must be a type of ASE Calculator')

        if labels is not None and len(labels) != self.length:
            raise ValueError('labels must be long as the collection itself')

        # Then set it up
        for i, s in enumerate(self.structures):
            if labels is None:
                # First: do we have a name?
                if 'name' in s.info:
                    label = s.info['name']
                else:
                    label = 'struct_{0}'.format(i)
            else:
                label = labels[i]
            calc = calctype(atoms=s,
                            label=str(label),
                            **params)
            # To make sure...
            s.set_calculator(calc)

    def run_calculators(self, properties=None, system_changes=None):
        """Run all previously set ASE calculators.

        | Args:
        |   properties (list[str]): list of properties to calculate (depends
        |                           on type of Calculator used)
        |   system_changes (list[str]): list of changes to the structure
        |                               since the last calculation. Can be
        |                               any combination of these five:
        |                               'positions', 'numbers', 'cell',
        |                               'pbc', 'initial_charges' and
        |                               'initial_magmoms'.

        """

        # First, check if we even have those
        if any([c is None for c in self.all.calc]):
            raise RuntimeError('Not all structures in collection'
                               ' have a calculator')
        kwargs = {}
        if properties is not None:
            kwargs['properties'] = properties
        if system_changes is not None:
            kwargs['system_changes'] = system_changes
        self.all.map(lambda s: s.calc.calculate(atoms=s, **kwargs))

    def chunkify(self, chunk_size=None, chunk_n=None):
        """Split this collection into multiple collections based on either
        size or number of chunks.

        | Args:
        |   chunk_size (Optional[int]): maximum size of a generated chunk
        |   chunk_n (Optional[int]): number of chunks to generate

        | Returns:
        |   chunks (list[AtomsCollection]): a list of the generated chunks

        """

        if [chunk_size, chunk_n].count(None) != 1:
            raise RuntimeError('Only one between chunk_size and chunk_n'
                               'must be passed')

        if chunk_size is not None and type(chunk_size) is not int:
            raise TypeError('chunk_size must be an int')
        if chunk_n is not None and type(chunk_n) is not int:
            raise TypeError('chunk_n must be an int')

        # Now determine the size
        if chunk_size is None:
            chunk_size = int(np.ceil(self.length*1.0/chunk_n))

        struct_chunks = [self.structures[i:i+chunk_size]
                         for i in range(0, self.length, chunk_size)]
        chunks = [AtomsCollection(s, self.info) for s in struct_chunks]

        # Now the arrays
        for aname in self._arrays:
            for i, c in enumerate(chunks):
                c.set_array(aname, self._arrays[aname][i*chunk_size:
                                                       i*(chunk_size+1)])

        return chunks

    def sorted_byarray(self, name, reverse=False):
        """Return a copy of this collection sorted by a given array.

        | Args:
        |   name (str): name of the array to use for the sorting
        |   reverse (Optional[bool]): reverse order of sorting (max to min)

        | Returns:
        |   sorted (AtomsCollection): a sorted copy of the collection

        """

        # First, check that we do have the array
        if not self.has(name):
            raise ValueError("Array \'{0}\' does not exist".format(name))

        arr_names = list(self._arrays.keys())
        data_block = zip(self.structures,
                         *[self._arrays[n] for n in arr_names])
        key_i = arr_names.index(name)
        data_block = list(zip(*sorted(data_block,
                                      key=lambda x: x[key_i+1],
                                      reverse=reverse)))

        sorted_copy = AtomsCollection(data_block[0], info=self.info)
        for ai, an in enumerate(arr_names):
            sorted_copy.set_array(an, data_block[ai+1])

        return sorted_copy

    def filter(self, filter_func):
        """Return a collection composed only of the elements for which a given
        filter function returns True.

        | Args:
        |   filter_func (function<Atoms>
        |                       => bool): filter function. Should take an
        |                                 Atoms object and return a boolean

        | Returns:
        |   filtered (AtomsCollection): the filtered version of the collection

        """

        filter_slice = []

        for i, s in enumerate(self.structures):
            if (filter_func(s)):
                filter_slice.append(i)

        return self[filter_slice]

    def classify(self, classes):
        """Return a dictionary of collections based on the names of assigned
        classes.

        | Args:
        |   classes (np.ndarray): array of the class to which each structure
        |                         belongs. For example [1, 2, 1] will put the
        |                         first and third structures in class 1 and
        |                         the other in class 2. The classes can be any
        |                         hashable types, like int or str.

        | Returns:
        |   classified (dict): a dictionary using class names as keys and 
        |                      sliced collections as values

        """

        classes = np.array(classes)
        classified = {k: self[np.where(classes == k)[0]]
                      for k in set(classes)}

        return classified

    def save(self, filename):
        """Simply save a pickled copy to a given file path"""

        # Pickling doesn't deal well with the _AllCaller, so we get rid of it
        selfcopy = self[:]
        selfcopy._all = None

        with open(filename, 'wb') as f:
            pickle.dump(selfcopy, f, protocol=2)

    @staticmethod
    def load(filename):
        """Load a pickled copy from a given file path"""

        with open(filename, 'rb') as f:
            coll = pickle.load(f)

        if not isinstance(coll, AtomsCollection):
            raise ValueError('File does not contain an AtomsCollection'
                             ' object')
        # Restore the _AllCaller
        coll._all = _AllCaller(coll.structures, ase.Atoms)
        return coll

    @staticmethod
    def check_tree(path):
        """Checks if a path is a valid 'tree' format for a collection. This is
        any folder that satisfies the following conditions:

        - contains a .collection file storing metadata
        - contains a series of folders matching the list stored in the
          .collection file, and nothing else

        This function will return 0 if both conditions are satisfied, 1 if 
        only the first is, 2 if no .collection file is found, and -1 if the
        folder itself doesn't exist.

        | Args: 
        |   path (str): path to check for whether it matches or not the 
        |               collection pattern

        | Returns:
        |   result (int): 0, 1 or 2 depending on the outcome of the checks
        """

        if not os.path.exists(path):
            return -1

        # Begin by checking whether there is a .collection file
        try:
            with open(os.path.join(path, '.collection'), 'rb') as f:
                coll = pickle.load(f)
        except (IOError, UnicodeDecodeError):
            return 2  # No or invalid .collection file found

        # Check if the directories match
        dirlist = coll['dirlist']
        dirs = glob.glob(os.path.join(path, '*'))
        all_dirs = all([os.path.isdir(d) for d in dirs])
        dirs = [os.path.relpath(d, path) for d in dirs]

        # Are they even all directories?
        if (not all_dirs or set(dirlist) != set(dirs)):
            return 1

        return 0

    def save_tree(self, path, save_format, name_root='structure',
                  opt_args={}, safety_check=3):
        """Save the collection's structures as a series of folders, named like
        the structures, inside a given parent folder (that will be created if
        not present). Arrays and info are stored in a pickled .collection file
        which works as metadata for the whole directory tree.
        The files can be saved in a format of choice, or a function can be
        passed that will save them in a custom way. Only one collection can be
        saved per folder.

        | Args:
        |   path (str): folder path in which the collection should be saved.
        |   save_format (str or function): format in which the structures
        |                                  should be saved.
        |                                  If a string, it will be used as a
        |                                  file extension. If a function, it
        |                                  must take as arguments the
        |                                  structure (an ase.Atoms object)
        |                                  the save path (a string), and any
        |                                  additional arguments passed as
        |                                  opt_args, and take care of saving
        |                                  the required files.
        |   name_root (str): name prefix to be used for structures when a name
        |                    is not available in their info dictionary
        |   opt_args (dict): dictionary of additional arguments to pass to
        |                    either ase.io.write (if save_format is a string)
        |                    or to the save_format function.
        |   safety_check (int): how much care should be taken not to overwrite
        |                       potentially important data in path. Can be a
        |                       number from 0 to 3.
        |                       Here's the meaning of the codes:
        |
        |                       3 (default): always ask before overwriting an
        |                         existing folder that passes the check_tree
        |                         control, raise an exception otherwise;
        |                       2: overwite any folder that passes fully the
        |                          check_tree control, raise an exception
        |                          otherwise;
        |                       1: overwrite any folder that passes fully the
        |                          check_tree control, ask for user input
        |                          otherwise;
        |                       0 (DANGER - use at your own risk!): no checks,
        |                         always overwrite path.

        """

        check = AtomsCollection.check_tree(path)

        def ow_ask(path):
            return utils.safe_input(('Folder {0} exists, '
                                     'overwrite (y/n)?').format(path)
                                    ).lower() == 'y'

        if check > -1:
            # The folder exists
            if check == 0:
                if safety_check >= 3:
                    # Ask for permission
                    perm = ow_ask(path)
                else:
                    perm = True
            else:
                if safety_check >= 2:
                    raise IOError(('Trying to overwrite folder {0} which did'
                                   ' not pass check_tree control (result {1})'
                                   ' with safety_check level '
                                   '{2}').format(path,
                                                 check,
                                                 safety_check))
                elif safety_check == 1:
                    perm = ow_ask(path)
                else:
                    perm = True

            if not perm:
                print('Can not overwrite folder {0}, skipping...'.format(path))

            shutil.rmtree(path)

        # Re-create folder
        os.mkdir(path)

        # Format type?
        is_ext = utils.is_string(save_format)
        is_func = hasattr(save_format, '__call__')
        if not (is_ext or is_func):
            raise ValueError('Invalid save_format passed to save_tree')

        dirlist = []
        for i, s in enumerate(self.structures):
            sname = s.info.get('name', '{0}_{1}'.format(name_root, i+1))
            fold = os.path.join(path, sname)
            try:
                os.mkdir(fold)
            except OSError:
                shutil.rmtree(fold)
                os.mkdir(fold)
            if is_ext:
                ase_io.write(os.path.join(fold, sname + '.' + save_format), s,
                             **opt_args)
            elif is_func:
                save_format(s, fold, **opt_args)

            dirlist.append(sname)

        with open(os.path.join(path, '.collection'), 'wb') as f:
            pickle.dump({'dirlist': dirlist,
                         'arrays': self._arrays,
                         'info': self.info}, f,
                        protocol=2)

    @staticmethod
    def load_tree(path, load_format, opt_args={}, safety_check=3,
                  tolerant=False):
        """Load a collection's structures from a series of folders, named like
        the structures, inside a given parent folder, as created by save_tree.
        The files can be loaded from a format of choice, or a
        function can be passed that will load them in a custom way.

        | Args:
        |   path (str): folder path in which the collection should be saved.
        |   load_format (str or function): format from which the structures
        |                                  should be loaded.
        |                                  If a string, it will be used as a
        |                                  file extension. If a function, it
        |                                  must take as arguments the load
        |                                  path (a string) and any additional
        |                                  arguments passed as opt_args, and
        |                                  return the loaded structure as an
        |                                  ase.Atoms object.
        |   opt_args(dict): dictionary of additional arguments to pass to
        |                   either ase.io.read (if load_format is a string)
        |                   or to the load_format function.
        |   safety_check (int): how much care should be taken to verify the
        |                       folder that is being loaded. Can be a number
        |                       from 0 to 3.
        |                       Here's the meaning of the codes:
        |
        |                       3 (default): only load a folder if it passes
        |                         fully the check_tree control;
        |                       2: load any folder that has a valid
        |                          .collection file, but only the listed
        |                          subfolders;
        |                       1: load any folder that has a valid
        |                          .collection file, all subfolders. Array
        |                          data will be discarded;
        |                       0: no checks, try to load from all subfolders.
        |   tolerant (bool): if set to true, proceeds to load the
        |                    structures into an AtomsCollection, even
        |                    if some of the structures could not be
        |                    read.

        | Returns:
        |   coll (AtomsCollection): loaded collection

        """

        check = AtomsCollection.check_tree(path)

        if check == -1:
            raise IOError('Folder {0} does not exist'.format(path))

        dirlist = []
        if check < 2:
            with open(os.path.join(path, '.collection'), 'rb') as f:
                coll = pickle.load(f)
            if check == 1 and safety_check == 3:
                raise IOError(('Folder {0} is not a valid collection '
                               'tree').format(path))
            if safety_check >= 2:
                dirlist = coll['dirlist']
            else:
                dirlist = [os.path.relpath(d, path)
                           for d in glob.glob(os.path.join(path, '*')) if
                           os.path.isdir(d)]
        else:
            if safety_check > 0:
                raise IOError(('Folder {0} is not a valid collection '
                               'tree').format(path))
            dirlist = [os.path.relpath(d, path)
                       for d in glob.glob(os.path.join(path, '*')) if
                       os.path.isdir(d)]

        # Format type?
        is_ext = utils.is_string(load_format)
        is_func = hasattr(load_format, '__call__')
        if not (is_ext or is_func):
            raise ValueError('Invalid load_format passed to load_tree')

        structures = []
        for d in dirlist:
            try:
                if is_ext:
                    s = ase_io.read(os.path.join(path, d, d + '.' +
                                    load_format), **opt_args)
                elif is_func:
                    s = load_format(os.path.join(path, d), **opt_args)
                structures.append(s)
            except Exception as e:
                warnings.warn(str(e))

        if check < 2:
            info = coll['info']
        else:
            info = {}

        percentage_failed = (1-len(structures)/len(dirlist))*100

        if percentage_failed > 0:

            if percentage_failed == 100:
                raise IOError("{0:.2f}% of structures could not be loaded.".
                              format(percentage_failed))
                return
            elif not tolerant:
                raise IOError("{0:.2f}% of structures could not be loaded. Set"
                              " tolerant to True if you would still"
                              " like to load the remaining structures."
                              .format(percentage_failed))
                return
            else:
                percentage_success = 100-percentage_failed
                info['percentage_loaded'] = percentage_success

                warnings.warn("{0:.2f}% of structures could not be loaded."
                              .format(percentage_failed))
        else:
            print("100% of structures loaded successfully.")

        loaded_coll = AtomsCollection(structures, info=info)

        if safety_check >= 2:
            arrays = coll['arrays']
            for k, a in arrays.items():
                loaded_coll.set_array(k, a)

        return loaded_coll


if __name__ == '__main__':

    # Just load whatever was passed in the command line!

    if len(sys.argv) > 1:
        coll = AtomsCollection(sys.argv[1:], progress=True, cell_reduce=True)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
optional.py

Contains imports for libraries considered optional, and decorators to include
checks for them.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from functools import wraps

try:
    import networkx as _networkx
except ImportError:
    _networkx = None

try:
    import spglib as _spglib
except ImportError:
    try:
        from pyspglib import spglib as _spglib
    except ImportError:
        _spglib = None

try:
    import sklearn as _sklearn
except ImportError:
    _sklearn = None


"""
These decorators check if the required module is available, if not print
an error message, if yes pass it as a variable to the function itself. 
They all take the name one desires the library to have within the function
as an argument. The function needs to have a named variable of the same name
in its interface.
"""


def requireNetworkX(import_name='networkx'):

    def decorator(func):

        @wraps(func)
        def wrapper(*args, **kwargs):

            if _networkx is None:
                raise RuntimeError('This function requires an installation of'
                                   ' NetworkX to work - please install it '
                                   'with:\n\tpip install networkx')
            else:
                kwargs[import_name] = _networkx
                return func(*args, **kwargs)

        return wrapper

    return decorator


def requireSpglib(import_name='spglib'):

    def decorator(func):

        @wraps(func)
        def wrapper(*args, **kwargs):

            v = list(map(int, _spglib.__version__.split('.')))

            if _spglib is None:
                raise RuntimeError('This function requires an installation of'
                                   ' spglib to work - please install it '
                                   'with:\n\tpip install spglib')
            elif v[0] < 1 or (v[0] == 1 and v[1] <= 8):
                raise RuntimeError('This function requires a version of'
                                   ' spglib superior to 1.8 to work - '
                                   'please install it '
                                   'with:\n\tpip install --upgrade spglib')
            else:
                kwargs[import_name] = _spglib
                return func(*args, **kwargs)

        return wrapper

    return decorator


def requireScikitLearn(import_name='sklearn'):

    def decorator(func):

        @wraps(func)
        def wrapper(*args, **kwargs):

            if _sklearn is None:
                raise RuntimeError('This function requires an installation of'
                                   ' scikit-learn to work - please install '
                                   'it with:\n\tpip install scikit-learn')
            else:
                kwargs[import_name] = _sklearn
                return func(*args, **kwargs)

        return wrapper

    return decorator



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes and functions for simulating X-ray diffraction
spectroscopic results from structures.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
import numpy as np
from collections import namedtuple
# Internal imports
from soprano import utils
from soprano.calculate.xrd import sel_rules as xrdsel
try:
    try:
        import spglib
    except ImportError:
        from pyspglib import spglib
except ImportError:
    raise ImportError('pySPGLIB not installed on this system, XRD'
                      ' functionality cannot be used.')


XraySpectrum = namedtuple("XraySpectrum", ("theta2", "hkl", "hkl_unique",
                                           "invd", "intensity", "lambdax"))

XraySpectrumData = namedtuple("XraySpectrumData", ("theta2", "intensity"))


class XRDCalculator(object):

    """A class implementing methods for XRD simulations, comparisons and
    fittings.

    """

    def __init__(self, lambdax=1.54056, theta2_digits=6, baseline=0.0,
                 peak_func=None, peak_f_args=None):
        """
        Initialize the XDRCalculator object's main parameters

        | Args:
        |    lambdax (Optional[float]): X-ray wavelength in Angstroms
        |                               (default is 1.54056 Ang)
        |    theta2_digits (Optional[int]): Rounding within which
        |                                  two theta angles (in degrees)
        |                                  are considered to be equivalent
        |                                  (default is 6 digits) when
        |                                  calculating theoretical peaks
        |   baseline (Optional[float]): baseline to use as starting point for
        |                               simulated spectra
        |   peak_func (Optional[function<float, float, \*kargs>
        |                       => <np.ndarray>]): the function used to
        |                                          simulate peaks. Should take
        |                                          th2 as its first argument,
        |                                          peak centre as its second,
        |                                          and any number of optional
        |                                          arguments. Returns a numpy
        |                                          array containing the peak
        |                                          shape. Should be able to
        |                                          work with numpy arrays as
        |                                          input
        |   peak_f_args (Optional[list<float>]): optional arguments for
        |                                        peak_func. If no peak_func
        |                                        has been supplied by the
        |                                        user, the first value will
        |                                        be used as the Gaussian width

        """

        # These properties are basic and can be changed by the user directly
        self.lambdax = lambdax
        self.theta2_digits = 6
        self.baseline = baseline
        # These ones not so much and need checking
        self.set_peak_func(peak_func, peak_f_args)

    @property
    def peak_func(self):
        """The function used to build peaks in simulated spectra

           Should be of form peak_func(theta2, peak_position, \*peak_f_args)

        """
        return self._peak_func

    @property
    def peak_f_args(self):
        """Additional arguments to be passed to peak_func"""
        return self._peak_f_args

    # The setter isn't part of the property because the option to pass None
    # has to be made clear as well
    def set_peak_func(self, peak_func=None, peak_f_args=None):
        """Set a new peak_func for this XDRCalculator. If no new function is
           passed, reset the default Gaussian function.

           | Args:
           |   peak_func (Optional[function<float, float, \*kargs>
           |                       => <np.ndarray>]): the function used to
           |                                          simulate peaks. Should
           |                                          take th2 as its first
           |                                          argument, peak centre as
           |                                          its second, and any
           |                                          number of optional
           |                                          arguments. Returns a
           |                                          numpy array containing
           |                                          the peak shape. Should
           |                                          be able to work with
           |                                          numpy arrays as input
           |   peak_f_args (Optional[list<float>]): optional arguments for
           |                                        peak_func. If no peak_func
           |                                        has been supplied by the
           |                                        user, the first value will
           |                                        be used as the Gaussian
           |                                        width
        """

        # If there's no peak function args...
        if peak_f_args is None:
            if peak_func is not None:
                peak_f_args = []
            else:
                peak_f_args = [0.1]  # Default Gaussian width

        # For peak_func there's little we can do to check, mostly it's the
        # user's responsibility.
        if peak_func is not None:
            nargs, nargs_def = utils.inspect_args(peak_func)
            if nargs < 2:
                raise ValueError("Invalid peak_func passed to set_peak_func")
            elif (nargs-(2+nargs_def)) > len(peak_f_args):
                raise ValueError("""Invalid number of peak_f_args passed to
                                    set_peak_func""")
        else:
            # A gaussian here
            peak_func = _gauss_peak_default
            if peak_f_args is None:
                # Use default width
                peak_f_args = [0.1]
            else:
                peak_f_args = peak_f_args[:1]  # Ignore all args but one

        self._peak_func = peak_func
        self._peak_f_args = peak_f_args

    def powder_peaks(self, atoms=None, latt_abc=None, n=1, o='all'):
        """
        Calculate the peaks (without intensities) of a powder
        XRD spectrum given either an Atoms object or the lattice in ABC form
        and the spacegroup indices to apply the selection rules

        | Args:
        |    atoms (Optional[soprano.Atoms]): atoms object to gather lattice
        |                                     and spacegroup information from
        |    latt_abc (Optional[np.ndarray]): periodic lattice in ABC form,
        |                                     Angstroms and radians
        |   n (Optional[int]): International number of the required spacegroup
        |   o (Optional[int]): Sub-option of the required spacegroup

        | Returns:
        |    xpeaks (XraySpectrum): a named tuple containing the peaks
        |                           with theta2, corresponding hkl indices,
        |                           a unique hkl tuple for each peak,
        |                           inverse reciprocal lattice distances,
        |                           intensities and wavelength

        | Raises:
        |   ValueError: if some of the arguments are invalid

        """

        # First a sanity check
        if sum(arg is not None for arg in [atoms, latt_abc]) != 1:
            raise ValueError("""One and only one between atoms and latt_abc
                                must be passed to powder_peaks""")
        if atoms is not None:
            # Define the lattice
            latt_abc = utils.cart2abc(atoms.get_cell())
            # And the symmetry
            symm_data = spglib.get_symmetry_dataset(atoms)
            h = int(symm_data['hall_number'])
            try:
                sel_rule = xrdsel.get_sel_rule_from_hall(h)
            except ValueError:
                # Not found?
                raise RuntimeWarning("""Required symmetry not found -
                                        No selection rules will be applied to
                                        XRD powder spectrum""")
        else:
            latt_abc = np.array(latt_abc, copy=False)
            if latt_abc.shape != (2, 3):
                raise ValueError(
                    "Invalid argument latt_abc passed to xrd_pwd_peaks")
            try:
                sel_rule = xrdsel.get_sel_rule_from_international(n, o)
            except ValueError:
                # Not found?
                raise RuntimeWarning("""Required symmetry not found -
                                        No selection rules will be applied to
                                        XRD powder spectrum""")

        inv_d_max = 2.0/self.lambdax  # Upper limit to the inverse distance

        # Second, find the matrix linking hkl indices to the inverse distance
        hkl2d2 = utils.hkl2d2_matgen(latt_abc)
        hkl_shape = utils.minimum_supcell(inv_d_max, r_matrix=hkl2d2)

        min_hkl_bounds = (-((np.array(hkl_shape)-1)/2)).astype(int)
        max_hkl_bounds = hkl_shape+min_hkl_bounds

        hrange = range(min_hkl_bounds[0], max_hkl_bounds[0])
        krange = range(min_hkl_bounds[1], max_hkl_bounds[1])
        lrange = range(min_hkl_bounds[2], max_hkl_bounds[2])

        # We now build a full grid of hkl indices. In this way
        # iteration is performed over numpy arrays and thus faster
        hkl_grid = np.array(
            np.meshgrid(hrange, krange, lrange)).reshape((3, -1))
        # First applying the selection rule
        selected_i = np.where(np.apply_along_axis(sel_rule, 0, hkl_grid))[0]
        hkl_grid = hkl_grid[:, selected_i]
        inv_d_grid = np.sqrt(np.sum(hkl_grid *
                                    np.tensordot(hkl2d2, hkl_grid,
                                                 axes=((1,), (0,))), axis=0))

        # Some will still have inv_d > inv_d_max, we fix that here
        # We also eliminate both 2theta = 0 and 2theta = pi to avoid
        # divergence later in the geometric factor
        valid_i = np.where((inv_d_grid < inv_d_max)*(inv_d_grid > 0))[0]
        hkl_grid = hkl_grid[:, valid_i]
        inv_d_grid = inv_d_grid[valid_i]
        theta_grid = np.arcsin(inv_d_grid/inv_d_max)
        # Now we calculate theta2.
        # Theta needs to be truncated to a certain number of
        # digits to avoid the same peak to be wrongly
        # considered as two.
        # This function also takes care of the sorting
        unique_sorting = np.unique(np.round(2.0*theta_grid*180.0/np.pi,
                                            self.theta2_digits),
                                   return_index=True,
                                   return_inverse=True)

        peak_n = len(unique_sorting[0])
        hkl_sorted = [hkl_grid[:,
                               np.where(unique_sorting[2] == i)[0]].T.tolist()
                      for i in range(peak_n)]
        hkl_unique = hkl_grid[:, unique_sorting[1]].T
        invd = inv_d_grid[unique_sorting[1]]
        xpeaks = XraySpectrum(unique_sorting[0],
                              np.array(hkl_sorted, dtype=object),
                              hkl_unique,
                              invd,
                              np.ones(peak_n),
                              self.lambdax)

        return xpeaks

    def exp_dataset(self, th2_axis, int_axis):
        """Build an experimental dataset as an XraySpectrumData object.

        | Args:
        |   th2_axis (np.ndarray): array containing the values for 2*theta
        |   int_axis (np.ndarray): array containing the values for intensity

        | Returns:
        |   exp_spec (XraySpectrumData): named tuple containing the
        |                                experimental dataset

        | Raises:
        |   ValueError: if some of the values passed are invalid

        """

        # Sanity checks as usual
        th2_axis = np.array(th2_axis, copy=False)
        int_axis = np.array(int_axis, copy=False)

        if th2_axis.shape != int_axis.shape or len(th2_axis.shape) != 1:
            raise ValueError("Invalid dataset passed to xrd_exp_dataset")

        return XraySpectrumData(th2_axis, int_axis)

    def dataset_range(self, xpeaks, theta2_range=(None, None)):
        """Restrict the given dataset (XraySpectrum or XraySpectrumData) to
           only the values that lie within a certain theta2 range.

        | Args:
        |   xpeaks (XraySpectrum or XraySpectrumData): the dataset to modify
        |   theta2_range (tuple<int>): a tuple indicating minimum and maximum
        |                              of the desired theta2 range (degrees).
        |                              A value of None means no boundary

        | Returns:
        |   xpeaks_restrict (XraySpectrum or XraySpectrumData): the restricted
        |                                                       dataset

        | Raises:
        |   ValueError: if some of the values passed are invalid

        """

        if type(xpeaks) not in (XraySpectrum, XraySpectrumData):
            raise ValueError("Invalid xpeaks passed to dataset_range")
        try:
            th2_min, th2_max = theta2_range
        except (TypeError, ValueError):
            raise ValueError("Invalid theta2_range passed to dataset_range")

        # Find the good indices
        th2_min = -np.inf if th2_min is None else th2_min
        th2_max = np.inf if th2_max is None else th2_max

        range_i = np.where((th2_min <= xpeaks.theta2) &
                           (xpeaks.theta2 <= th2_max))[0]

        # Switch type
        if type(xpeaks) is XraySpectrum:
            return XraySpectrum(xpeaks.theta2[range_i],
                                xpeaks.hkl[range_i],
                                xpeaks.hkl_unique[range_i],
                                xpeaks.invd[range_i],
                                xpeaks.intensity[range_i],
                                xpeaks.lambdax)
        else:
            return XraySpectrumData(xpeaks.theta2[range_i],
                                    xpeaks.intensity[range_i])

    def spec_simul(self, xpeaks, th2_axis):
        """
        Simulate an XRD spectrum given positions of peaks, intensities,
        baseline, and a peak function (a Gaussian by default).

        | Args:
        |   xpeaks (XraySpectrum): object containing the details of the XRD
        |                          peaks
        |   th2_axis (np.ndarray): theta2 axis points on which the
        |                          spectrum should be simulated

        | Returns:
        |   simul_spec (XraySpectrumData): simulated XRD spectrum
        |   simul_peaks (np.ndarray): simulated spectrum intensities broken by
        |                             peak contribution along axis 1

        | Raises:
        |   ValueError: if some of the arguments are invalid

        """

        # Sanity checks here

        peaks_th2 = np.array(xpeaks.theta2, copy=False)
        peaks_int = np.array(xpeaks.intensity, copy=False)
        th2_axis = np.array(th2_axis, copy=False)
        if len(th2_axis.shape) != 1:
            raise ValueError("Invalid th2_axis passed to spec_simul")

        # So here we are. Actual simulation!
        simul_peaks = peaks_int[None, :]*self.peak_func(th2_axis[:, None],
                                                        peaks_th2[None, :],
                                                        *self.peak_f_args)

        simul_spec = np.sum(simul_peaks, axis=1) + \
            np.ones(len(th2_axis))*self.baseline
        simul_spec = XraySpectrumData(th2_axis, simul_spec)

        return simul_spec, simul_peaks

    def lebail_fit(self, xpeaks, exp_spec,
                   rwp_tol=1e-2, max_iter=100):
        """
        Perform a refining on an XraySpectrum object's intensities based on
        experimental data with leBail's method.

        | Args:
        |   xpeaks (XraySpectrum): object containing the details of the XRD
        |                          peaks
        |   exp_spec (XraySpectrumData): experimental data, dataset built
        |                                using xrd_exp_dataset
        |   rwp_tol (Optional[float]): tolerance on the Rwp error value
        |                              between two iterations that stops the
        |                              calculation. Default is 1e-2
        |   max_iter (Optional[int]): maximum number of iterations to perform

        | Returns:
        |   xpeaks_scaled (XraySpectrum): a new XraySpectrum object, with
        |                                 intensities properly scaled to match
        |                                 the experimental data
        |   simul_spec (np.ndarray): final simulated XRD spectrum
        |   simul_peaks (np.ndarray): final simulated spectrum broken by peak
        |                             contribution along axis 1
        |   rwp (float): the final value of Rwp (fitness of simulated to
        |                experimental data)

        | Raises:
        |   ValueError: if some of the arguments are invalid

        """

        # Sanity check (the rest is up to the other functions)
        if type(xpeaks) != XraySpectrum:
            raise ValueError("Invalid xpeaks passed to lebail_fit")
        if type(exp_spec) != XraySpectrumData:
            raise ValueError("Invalid exp_spec passed to lebail_fit")

        # Copy xpeaks so that the original object will stay unaltered
        xpeaks = copy.deepcopy(xpeaks)

        # Fetch the data
        peaks_th2 = xpeaks.theta2
        th2_axis = np.array(exp_spec.theta2, copy=False)
        int_axis = np.array(exp_spec.intensity, copy=False)
        # Find a suitable starting height for the peaks
        peaks_int = (xpeaks.intensity*0.0)+np.amax(int_axis)/4.0

        # Perform the first simulation
        simul_spec, simul_peaks = self.spec_simul(xpeaks, th2_axis)
        # Calculate Rwp
        rwp0 = _Rwp_eval(simul_spec.intensity, int_axis)

        # Now the iterations
        for i in range(max_iter):
            # Perform a correction
            peaks_int *= _leBail_rescale_I(simul_peaks, simul_spec.intensity,
                                           int_axis)
            # Store
            np.copyto(xpeaks.intensity, peaks_int)
            # Recalculate
            simul_spec, simul_peaks = self.spec_simul(xpeaks, th2_axis)
            # Calculate Rwp
            rwp = _Rwp_eval(simul_spec.intensity, int_axis)
            # Stop if already zero, we can't get any better
            if (rwp == 0):
                break
            # Gain?
            delta_rwp = abs((rwp-rwp0)/rwp)
            if delta_rwp < rwp_tol:
                break
            rwp0 = rwp

        return xpeaks, simul_spec, simul_peaks, rwp


# Here be utility functions that are NOT meant to be used by the uninitiated.
# Tread at your own peril.


def _Rwp_eval(simul_int, exp_int):
    """Evaluate Rwp for use in LeBail fitting"""

    weights = 1.0/exp_int

    r_wp = np.sqrt(np.sum(weights*(exp_int-simul_int)**2.0) /
                   np.sum(weights*exp_int**2.0))

    return r_wp


def _leBail_rescale_I(simul_peaks, simul_spec, exp_spec):
    """Returns rescaling factors for intensities in leBail fitting"""

    sum_peaks = np.sum(simul_peaks, axis=0)
    # It could be zero somewhere, fix that
    sum_peaks = np.where(sum_peaks > 0, sum_peaks, np.inf)
    i_scale = np.sum(exp_spec[:, None]*simul_peaks/simul_spec[:, None],
                     axis=0)/sum_peaks

    return i_scale


def _gauss_peak_default(x, x0, w):
    """Gaussian peak function (for spectrum simulation)"""
    return np.exp(-((x-x0)/w)**2.0)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of AtomsProperties that relate to labeling of systems"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from soprano.properties import AtomsProperty
from soprano.utils import recursive_mol_label
from soprano.properties.linkage import Molecules, HydrogenBonds, Bonds


class SiteLabels(AtomsProperty):

    """
    SiteLabels

    Compute a unique label for an atom belonging to a molecule by 
    exploiting network topology. Atoms can have the same label, but only if
    they're fundamentally indistinguishable in the molecule's chemical context
    (for example, three hydrogen atoms on a CH3 group). The label describes 
    the molecular network as traversed starting from the given site, with each
    pair of square brackets indicating the traversal of a further bond, and 
    all paths kept to the shortest possible.

    | Parameters:
    |   force_recalc (bool): if True, always recalculate the molecules even if
    |                        already present.
    |   sites (list[int]): the sites for which the label have to be calculated.
    |                      If not given, all of them are calculated.
    |   custom_symbol (str): if present, replace the usual chemical symbol
    |                        with a custom name for the site on which the
    |                        label is being calculated, to make it
    |                        distinguishable from equivalent elements.

    | Returns:
    |   site_labels (list[str]): A list of the computed site labels
    """

    default_name = 'site_labels'
    default_params = {
        'force_recalc': False,
        'sites': None,
        'custom_symbol': None
    }

    @staticmethod
    def extract(s, force_recalc, sites, custom_symbol):

        # First, we need the molecules
        if Molecules.default_name not in s.info or force_recalc:
            Molecules.get(s)
        mols = s.info[Molecules.default_name]

        elems = s.get_chemical_symbols()

        if sites is None:
            sites = range(len(s))

        labels = []

        for s_i in sites:
            # Find the molecule that it belongs to
            s_mol = [m for m in mols if s_i in m.indices][0]
            s_elems = np.array(elems, dtype='S2')
            if custom_symbol is not None:
                s_elems[s_i] = custom_symbol
            # Grab the bonds
            bonds = s_mol.get_array(Bonds.default_name)
            # This is a necessary step since the bonds are not classified
            # by original structure index yet
            bonds = {a: bonds[i] for i, a in enumerate(s_mol.indices)}
            labels.append(recursive_mol_label(s_i, s_mol.indices, 
                                              bonds, s_elems))

        return labels


class MoleculeSites(AtomsProperty):

    """
    MoleculeSites

    Assigns unique labels to atoms belonging to molecules by exploiting
    network topology. Atoms can have the same label, but only if they're
    fundamentally indistinguishable in the molecule's chemical context
    (for example, three hydrogen atoms on a CH3 group). The molecule will be
    described by a characteristic string and by a series of labels in the
    format [element]_[number]. These sites will be saved by default and can
    be used for better insight when carrying out other analysis.

    | Parameters:
    |   force_recalc (bool): if True, always recalculate the molecules even if
    |                        already present.
    |   save_info (bool): if True, save the found molecular sites as part of
    |                     the Atoms object info. By default True.
    |   save_asarray (bool): if True the molecular site names are also saved
    |                        as an array of the molecule selection.

    | Returns:
    |   molecular_sites (dict): A dictionary containing info characterising
    |                           the molecule's chemical sites unequivocally.
    |                           These are a string representation of the
    |                           molecule itself and a dictionary linking
    |                           atomic indices (as found in the molecule in
    |                           AtomSelection form) to site labels.
    """

    default_name = 'molecule_sites'
    default_params = {
        'force_recalc': False,
        'save_info': True,
        'save_asarray': False
    }

    @staticmethod
    def extract(s, force_recalc, save_info, save_asarray):

        # First, we need the molecules
        if Molecules.default_name not in s.info or force_recalc:
            Molecules.get(s)

        elems = s.get_chemical_symbols()

        mol_sites = []

        for mol_i, mol in enumerate(s.info[Molecules.default_name]):

            # For each atom we do a depth-first traversal of the network
            sites = {}
            bonds = mol.get_array(Bonds.default_name)
            # This is a necessary step since the bonds are not classified
            # by original structure index yet
            bonds = {a: bonds[i] for i, a in enumerate(mol.indices)}
            for a in mol.indices:
                sites[a] = recursive_mol_label(a, mol.indices, bonds, elems)

            # Now grab the unique sites and pick the name of the molecule
            site_names = sorted(list(set(sites.values())))
            site_dict = {'name': site_names[0]}
            # Now rename the sites
            elem_sites = {}
            for a in sites:
                s_i = site_names.index(sites[a])
                if elems[a] not in elem_sites:
                    elem_sites[elems[a]] = [s_i]
                elif s_i not in elem_sites[elems[a]]:
                    elem_sites[elems[a]].append(s_i)
                sites[a] = '{0}_{1}'.format(elems[a],
                                            elem_sites[elems[a]].index(s_i)+1)

            site_dict['sites'] = sites
            mol_sites.append(site_dict)

            if save_asarray:
                arr = [sites[a] for a in mol.indices]
                mol.set_array(MoleculeSites.default_name, arr)

        if save_info:
            s.info[MoleculeSites.default_name] = mol_sites

        return mol_sites


class HydrogenBondTypes(AtomsProperty):

    """
    HydrogenBondTypes

    Assign MoleculeSites labels to atoms, then characterise existing hydrogen
    bonds based on them, and return a list of such bonds detected in a system.
    The bonds come in the form '{0}<{1},{2}>..{3}<{4}>', where {0} is the name
    of the molecule containing the hydrogen, {2} is the hydrogen, {1} the atom
    to which the hydrogen is bonded, {3} the name of the other molecule and
    {4} the atom to which the hydrogen is hydrogen bonded.

    | Parameters:
    |   force_recalc (bool): if True, always recalculate the molecules even if
    |                        already present.
    |   save_info (bool): if True, save the found hydrogen bond types as part
    |                     of the Atoms object info. By default True.

    | Returns:
    |   hydrogen_bond_types (list): A list containing info characterising the
    |                               hydrogen bonds present in the system in a
    |                               detailed way.
    """

    default_name = 'hydrogen_bond_types'
    default_params = {
        'force_recalc': False,
        'save_info': True,
    }

    @staticmethod
    def extract(s, force_recalc, save_info):

        # First, we need the molecules
        if Molecules.default_name not in s.info or force_recalc:
            Molecules.get(s)

        # Then the hydrogen bonds
        if HydrogenBonds.default_name not in s.info or force_recalc:
            HydrogenBonds.get(s)

        # Finally the sites
        if MoleculeSites.default_name not in s.info or force_recalc:
            MoleculeSites.get(s)

        mols = s.info[Molecules.default_name]
        hbonds = s.info[HydrogenBonds.default_name]
        all_sites = s.info[MoleculeSites.default_name]

        hblabels = []
        for hbtype in hbonds:
            for hb in hbonds[hbtype]:
                A_i = hb['A'][0]
                H_i = hb['H']
                B_i = hb['B'][0]

                # Check in which molecule they are
                AH_sites = None
                B_sites = None
                for m_i, m in enumerate(mols):
                    if A_i in m:
                        AH_sites = all_sites[m_i]
                    if B_i in m:
                        B_sites = all_sites[m_i]

                if AH_sites is None or B_sites is None:
                    raise RuntimeError('Invalid hydrogen bond detected')

                # Now build the proper definition
                hblabel = ('{0}<{1},{2}>'
                           '..{3}<{4}>').format(AH_sites['name'],
                                                AH_sites['sites'][A_i],
                                                AH_sites['sites'][H_i],
                                                B_sites['name'],
                                                B_sites['sites'][B_i])
                hblabels.append(hblabel)

        return sorted(hblabels)


class CarbonHybridationState(AtomsProperty):

    """
    CarbonHybridationState

    Returns an integer indicating the hybridation state of any given carbon
    atom in a system, 1, 2, or 3 for sp1, sp2 or sp3, obtained by counting its
    bonds. For non-carbon atoms or carbon atoms that can't be identified
    returns 0.

    | Parameters:
    |   bonds ([tuple]): if present, use this list of bonds instead of
    |                    recalculating them (must be formatted like the output
    |                    of the linkage.Bonds property).
    |   save_info (bool): if True, save the found hybridation states as an
    |                     array in the Atoms object. By default True.

    | Returns:
    |   hybrid ([int]): list of integers encoding the hybridation states

    """

    default_name = 'carbon_hybridation_state'
    default_params = {
        'bonds': None,
        'save_info': True
    }

    @staticmethod
    def extract(s, bonds, save_info):

        elems = np.array(s.get_chemical_symbols())
        hybrid = np.zeros(len(elems)).astype(int)

        if (elems == 'C').any():
            # Only do this if there is any carbon...

            if bonds is None:
                # Recalculate bonds
                bonds = Bonds.get(s)

            C_i = set(list(np.where(elems == 'C')[0]))

            for b in bonds:
                C_b = C_i.intersection(b[:2])
                hybrid[list(C_b)] += 1

            # Now go from number of bonds to hybridation state
            hybrid[list(C_i)] = np.where((hybrid[list(C_i)] > 1) *
                                         (hybrid[list(C_i)] <= 4),
                                         hybrid[list(C_i)]-1,
                                         0)

        if save_info:
            s.set_array(CarbonHybridationState.default_name, hybrid)

        return hybrid



#!/usr/bin/env python
"""
Soprano - A library for cracking crystals!
by Simone Sturniolo
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# The next line is removed because it causes issues in interpreting
# the package_data line, unfortunately
# from __future__ import unicode_literals

from setuptools import setup, find_packages
from soprano import __version__

long_description = """
Soprano is a Python library developed and maintained by the CCP for NMR
Crystallography as a tool to help scientists working with crystallography and
simulations to generate, manipulate, run calculations on and analyse large
data sets of crystal structures, with a particular attention to the output of
ab-initio random structure searching,
or AIRSS. (https://www.mtg.msm.cam.ac.uk/Codes/AIRSS)

It provides a number of functionalities to help automate many common tasks
in computational crystallography."""

if __name__ == '__main__':

    setup(name='Soprano',
          version=__version__,
          description='A Python library to crack crystals',
          long_description=long_description,
          url='https://ccp-nc.github.io/soprano/',
          author='Simone Sturniolo',
          author_email='simone.sturniolo@stfc.ac.uk',
          license='LGPL',
          classifiers=[
              # How mature is this project? Common values are
              #   3 - Alpha
              #   4 - Beta
              #   5 - Production/Stable
              'Development Status :: 4 - Beta',

              # Indicate who your project is intended for
              'Intended Audience :: Science/Research',
              'Topic :: Scientific/Engineering :: Chemistry',
              'Topic :: Scientific/Engineering :: Physics',
              'Topic :: Scientific/Engineering :: Information Analysis',

              # Pick your license as you wish (should match "license" above)
              'License :: OSI Approved :: GNU Library or Lesser General Public'
              ' License (LGPL)',

              # Specify the Python versions you support here. In particular,
              # ensure that you indicate whether you support Python 2, Python 3
              # or both.
              'Programming Language :: Python :: 3',
          ],
          keywords=['crystallography', 'ccpnc', 'computational chemistry'],
          packages=find_packages(),
          # For data files. Example: 'soprano': ['data/*.json']
          package_data={'soprano': ['data/*.json']},
          entry_points={
              'console_scripts': ['vasp2cell = '
                                  'soprano.scripts.vasp2cell:__main__',
                                  'soprano_submitter = '
                                  'soprano.hpc.submitter:__main__',
                                  'phylogen = '
                                  'soprano.scripts.phylogen:__main__',
                                  'magresaverage = '
                                  'soprano.scripts.msaverage:__main__']
          },
          use_2to3=False,
          convert_2to3_doctests=[],
          # Requirements
          install_requires=[
              'numpy',
              'scipy',
              'ase'
          ],
          python_requires='>=3.1.*'
          )



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generator producing structures rattled of a given amount"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import numpy as np
# Internal imports
import soprano.utils as utils


def rattleGen(struct, amplitude=0.01, n=100, method='uniform'):
    """Generator function to create multiple structures by randomly displacing
    atoms of a given amount.

    | Args:
    |   struct (ase.Atoms): the starting structure to randomize
    |   amplitude (float or np.ndarray): the amplitude of the random
    |                                    displacement. Can be a single float
    |                                    for all atoms, a 1D numpy array of
    |                                    length N (N being the number of
    |                                    atoms, one value each) or a 2D numpy
    |                                    array of shape (N,3) (one value for
    |                                    each dimension).
    |                                    These values are used as interval for
    |                                    uniform random numbers and as stdev
    |                                    for normal random numbers
    |   n (int): maximum number of structures to generate. If set to None will
    |            generate infinite structures
    |   method (str): must be either 'uniform' or 'normal'. In the first case
    |                 the rattling will be a uniform random number between
    |                 +amplitude and -amplitude. In the second case it will be
    |                 a gaussian random number with +amplitude standard
    |                 deviation.

    | Returns:
    |   rattleGenerator (generator): an iterator that yields copies of the
    |                                base structure with randomly displaced
    |                                atoms.

    """

    if method not in ('uniform', 'normal'):
        raise ValueError('Invalid method argument passed to rattleGen')

    pos = struct.get_positions()

    # Check amplitude shape
    try:
        amplitude = np.array(amplitude)*1.0
    except TypeError:
        raise ValueError('Invalid amplitude argument passed to rattleGen')

    ampsh = amplitude.shape
    if len(ampsh) == 1:
        if ampsh != (pos.shape[0],):
            raise ValueError('Shape mismatch between amplitude and struct '
                             'arguments passed to rattleGen')
        amplitude = amplitude[:, None]
    elif len(ampsh) == 2:
        if ampsh != pos.shape:
            raise ValueError('Shape mismatch between amplitude and struct '
                             'arguments passed to rattleGen')

    i = 0

    n = np.inf if n is None else n

    while i < n:

        rnds = struct.copy()

        # Rattle positions
        if method == 'uniform':
            dxyz = (np.random.random(pos.shape)-0.5)*2*amplitude
        elif method == 'normal':
            dxyz = np.random.normal(size=pos.shape)*amplitude

        rnds.set_positions(pos+dxyz)

        yield rnds
        i += 1



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Classes and functions to carry out calculations using the bindings to GULP
(General Utility Lattice Program), a software providing a lot of useful
calculations with empirical force fields, partial charge calculations, Ewald
summation of Coulombic interactions and more. GULP can be found at:

http://nanochemistry.curtin.edu.au/gulp/

It needs to be installed on your system to use any of the functionality
provided here.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import

from soprano.calculate.gulp.w99 import get_w99_energy, W99Error
from soprano.calculate.gulp.charges import get_gulp_charges



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Module containing  a special set of AtomsProperties that transform an Atoms
object into another (by translating, rotating or mirroring all or some ions,
and so on). These all accept an Atoms object and some parameters and return
an Atoms object as well. Default behaviour for the .get method in most cases
will be to do nothing at all, these properties are meant to be instantiated.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import

from soprano.properties.transform.transform import (Translate, Rotate, Mirror)
